% EV3 command construction
%
% Methods::
% DataStream                   Constructor, creates an empty command object
% delete                    Destructor, clears the command object
%
% addHeaderSystem           Adds a system header to the command object
% addHeaderSystemReply      Adds a system header with reply to the command object
% addHeaderDirect           Adds a direct header to the command object
% addHeaderDirectReply      Adds a direct header with reply to the command object
% addLength                 Adds the length of the msg to the front of the command object
% addSystemCommand          Adds a system command to the command object
% addDirectCommand          Adds a direct command to the command object
%
% checkForError @MMI        Checks error byte in received package
% isCorrupt     @MMI        Checks whether reply packet is corrupt
%
% clear                     Clears the command msg
% display                   Displays the command msg (decimal)
% displayHex                Displays the command msg (hex)
%
% LC0                       Adds a local constant 0 to the command object
% LC1                       Adds a local constant 1 to the command object
% LC2                       Adds a local constant 2 to the command object
% LC4                       Adds a local constant 4 to the command object
% LV0                       Adds a local variable 0 to the command object
% GV0                       Adds a global variable 0 to the command object
% LCS                       Adds a local constant string to the command object
%
% addValue                  Adds a numerical value to the command object
% addArray                  Adds a numerical array to the command object
% addString                 Adds a string to the command object
% addLCSString              Adds a LCS string to the command object
%
% LONGToBytes               Adds a LONGToBytes to the command object
% WORDToBytes               Adds a WORDToBytes to the command object
% BYTEToBytes               Adds a BYTEToBytes to the command object
%
% PROGRAMHeader             Adds a PROGRAMHeader to the command object
% addFileSize               Adds the filesize to the command object
% VMTHREADHeader            Adds a VMTHREADHeader to the command object
% SUBCALLHeader             Adds a SUBCALLHeader to the command object
% BLOCKHeader               Adds a BLOCKHeader to the command object
% GenerateByteCode          Prints the command message to a file
%
% opNOP                     Adds a opNOP opcode to the command object
% opOBJECT_END              Adds a opOBJECT_END opcode to the command object
%
% opJR                      Adds a opJR opcode to the command object
%
% opUI_FLUSH                Adds a opUI_FLUSH opcode to the command object
%
% opUI_READ_GET_VBATT       Adds a opUI_READ opcode with a GET_VBATT subcode to the command object
% opUI_READ_GET_LBATT       Adds a opUI_READ opcode with a GET_LBATT subcode to the command object
%
% opUI_WRITE_PUT_STRING     Adds a opUI_WRITE opcode with a PUT_STRING subcode to the command object
% opUI_WRITE_INIT_RUN       Adds a opUI_WRITE opcode with a INIT_RUN subcode to the command object
% opUI_WRITE_LED            Adds a opUI_WRITE opcode with a LED subcode to the command object
%
% opUI_DRAW_UPDATE          Adds a opUI_DRAW opcode with a UPDATE subcode to the command object
% opUI_DRAW_CLEAN           Adds a opUI_DRAW opcode with a CLEAN subcode to the command object
% opUI_DRAW_PIXEL           Adds a opUI_DRAW opcode with a PIXEL subcode to the command object
% opUI_DRAW_LINE            Adds a opUI_DRAW opcode with a LINE subcode to the command object
% opUI_DRAW_CIRCLE          Adds a opUI_DRAW opcode with a CIRCLE subcode to the command object
% opUI_DRAW_TEXT            Adds a opUI_DRAW opcode with a TEXT subcode to the command object
% opUI_DRAW_VALUE           Adds a opUI_DRAW opcode with a VALUE subcode to the command object
% opUI_DRAW_FILLRECT        Adds a opUI_DRAW opcode with a FILLRECT subcode to the command object
% opUI_DRAW_RECT            Adds a opUI_DRAW opcode with a RECT subcode to the command object
% opUI_DRAW_INVERSERECT     Adds a opUI_DRAW opcode with a INVERSERECT subcode to the command object
% opUI_DRAW_SELECT_FONT     Adds a opUI_DRAW opcode with a SELECT_FONT subcode to the command object
% opUI_DRAW_TOPLINE         Adds a opUI_DRAW opcode with a TOPLINE subcode to the command object
% opUI_DRAW_FILLWINDOW      Adds a opUI_DRAW opcode with a FILLWINDOW subcode to the command object
% opUI_DRAW_FILLCIRCLE      Adds a opUI_DRAW opcode with a FILLCIRCLE subcode to the command object
% opUI_DRAW_STORE           Adds a opUI_DRAW opcode with a STORE subcode to the command object
% opUI_DRAW_RESTORE         Adds a opUI_DRAW opcode with a RESTORE subcode to the command object
%
% opTIMER_WAIT              Adds a opTIMER opcode with a WAIT subcode to the command object
% opTIMER_READY             Adds a opTIMER opcode with a READY subcode to the command object
% opTIMER_READ              Adds a opTIMER opcode with a READ subcode to the command object
%
% opSOUND_BREAK             Adds a opSOUND opcode with a BREAK subcode to the command object
% opSOUND_TONE              Adds a opSOUND opcode with a TONE subcode to the command object
% opSOUND_PLAY              Adds a opSOUND opcode with a PLAY subcode to the command object
% opSOUND_REPEAT            Adds a opSOUND opcode with a REPEAT subcode to the command object
% opSOUND_TEST              Adds a opSOUND opcode with a TEST subcode to the command object
% opSOUND_READY             Adds a opSOUND opcode with a READY subcode to the command object
%
% opINPUT_DEVICE_LIST                   Adds a INPUT_DEVICE opcode with a DEVICE_LIST subcode to the command object
% opINPUT_DEVICE_GET_TYPEMODE           Adds a INPUT_DEVICE opcode with a GET_TYPEMODE subcode to the command object
% opINPUT_DEVICE_SET_TYPEMODE   @MMI:   Adds a INPUT_DEVICE opcode with a SET_TYPEMODE subcode to the command object
% opINPUT_DEVICE_GET_SYMBOL             Adds a INPUT_DEVICE opcode with a GET_SYMBOL subcode to the command object
% opINPUT_DEVICE_CLR_ALL                Adds a INPUT_DEVICE opcode with a CLR_ALL subcode to the command object
% opINPUT_DEVICE_CLR_CHANGES    @MMI:   Adds a INPUT_DEVICE opcode with a CLR_CHANGES subcode to the command object
% opINPUT_DEVICE_GET_NAME               Adds a INPUT_DEVICE opcode with a GET_NAME subcode to the command object
% opINPUT_DEVICE_GET_MODENAME   @MMI:   Adds a INPUT_DEVICE opcode with a GET_MODENAME subcode to the command object
% opINPUT_DEVICE_GET_CONNECTION @MMI:   Adds a INPUT_DEVICE opcode with GET_CONNECTION subcode to the command object
% opINPUT_DEVICE_GET_MINMAX     @MMI:   Adds a INPUT_DEVICE opcode with a GET_MINMAX subcode to the command object 
% opINPUT_DEVICE_GET_CHANGES    @MMI:   Adds a INPUT_DEVICE opcode with a GET_CHANGES subcode to the command object
% opINPUT_DEVICE_GET_FORMAT     @MMI:   Adds a INPUT_DEVICE opcode with a GET_FORMAT subcode to the command object
% opINPUT_DEVICE_GET_BUMPS      @MMI:   Adds a INPUT_DEVICE opcode with a GET_BUMPS subcode to the command object
% opINPUT_READY                         Adds a opINPUT_READY opcode to the command object
% opINPUT_TEST                  @MMI:   Adds a opINPUT_TEST opcode to the command object
% opINPUT_READ                          Adds a opINPUT_READ opcode to the command object
% opINPUT_READSI                        Adds a opINPUT_READSI opcode to the command object
%
% opOUTPUT_SET_TYPE         Adds a opOUTPUT_SET_TYPE opcode to the command object
% opOUTPUT_RESET            Adds a opOUTPUT_RESET opcode to the command object
% opOUTPUT_STOP             Adds a opOUTPUT_STOP opcode to the command object
% opOUTPUT_SPEED            Adds a opOUTPUT_SPEED opcode to the command object
% opOUTPUT_POWER            Adds a opOUTPUT_POWER opcode to the command object
% opOUTPUT_START            Adds a opOUTPUT_START opcode to the command object
% opOUTPUT_POLARITY         Adds a opOUTPUT_POLARITY opcode to the command object
% opOUTPUT_READ             Adds a opOUTPUT_READ opcode to the command object
% opOUTPUT_TEST             Adds a opOUTPUT_TEST opcode to the command object
% opOUTPUT_READY            Adds a opOUTPUT_READY opcode to the command object
% opOUTPUT_STEP_POWER       Adds a opOUTPUT_STEP_POWER opcode to the command object
% opOUTPUT_TIME_POWER @MMI: Adds a opOUTPUT_TIME_POWER opcode to the command object
% opOUTPUT_STEP_SPEED       Adds a opOUTPUT_STEP_SPEED opcode to the command object
% opOUTPUT_TIME_SPEED       Adds a opOUTPUT_TIME_SPEED opcode to the command object
% opOUTPUT_STEP_SYNC        Adds a opOUTPUT_STEP_SYNC opcode to the command object
% opOUTPUT_TIME_SYNC        Adds a opOUTPUT_TIME_SYNC opcode to the command object
% opOUTPUT_CLR_COUNT        Adds a opOUTPUT_CLR_COUNT opcode to the command object
% opOUTPUT_GET_COUNT        Adds a opOUTPUT_GET_COUNT opcode to the command object
%
% opCOM_TEST  @MMI:         Adds a opCOM_TEST opcode to the command object    
% opCOM_READY @MMI:         Adds a opCOM_READY opcode to the command object
% opCOMGET_GET_BRICKNAME    Adds a opCOMGET opcode with a GET_BRICKNAME subcode to the command object
% opCOMSET_SET_BRICKNAME    Adds a opCOMSET opcode with a SET_BRICKNAME subcode to the command object
% opCOMGET_NETWORK @MMI:    Adds a opCOMGET opcode with a GET_NETWORK subcode to the command object
% opCOMGET_ID @MMI:         Adds a opCOMGET opcode with a GET_ID subcode to the command object
%
% opMAILBOX_WRITE           Adds a opMAILBOX_WRITE opcode to the command object
%
% BEGIN_DOWNLOAD            Adds a BEGIN_DOWNLOAD system command to the command object
% CONTINUE_DOWNLOAD         Adds a CONTINUE_DOWNLOAD system command to the command object
% BEGIN_UPLOAD              Adds a BEGIN_UPLOAD system command to the command object
% CONTINUE_UPLOAD           Adds a CONTINUE_UPLOAD system command to the command object
% LIST_FILES                Adds a LIST_FILES system command to the command object
% CONTINUE_LIST_FILES       Adds a CONTINUE_LIST_FILES system command to the command object
% CREATE_DIR                Adds a CREATE_DIR system command to the command object
% DELETE_FILE               Adds a DELETE_FILE system command to the command object
% WRITEMAILBOX              Adds a WRITEMAILBOX system command to the command object
% 
% Notes::
% - Refer to the EV3 documentation or source code for a more detailed
% description of the commands.
%
% Example::
%                   dataStream = DataStream();
%                   dataStream.addHeaderDirect(42,0,0);
%                   dataStream.opSOUND_TONE(volume,frequency,duration);
%                   dataStream.addLength();


classdef (Abstract) DataStream < handle
    % Communications format (c_com.h):
    % /*
    %       System Command Bytes:
    %   ,------,------,------,------,------,------,------,------,
    %   |Byte 0|Byte 1|Byte 2|Byte 3|Byte 4|Byte 5|      |Byte n|
    %   '------'------'------'------'------'------'------'------'
    % 
    %   Byte 0 - 1: Command size, Little Endian\n
    % 
    %   Byte 2 - 3: Message counter, Little Endian\n
    % 
    %   Byte 4:     Command type. See following defines   */
    % 
    %   #define     SYSTEM_COMMAND_REPLY          0x01    //  System command, reply required
    %   #define     SYSTEM_COMMAND_NO_REPLY       0x81    //  System command, reply not required
    % /*
    %   Byte 5:     System Command. see following defines */
    % 
    %   #define     BEGIN_DOWNLOAD                0x92    //  Begin file down load
    %   #define     CONTINUE_DOWNLOAD             0x93    //  Continue file down load
    %   #define     BEGIN_UPLOAD                  0x94    //  Begin file upload
    %   #define     CONTINUE_UPLOAD               0x95    //  Continue file upload
    %   #define     BEGIN_GETFILE                 0x96    //  Begin get bytes from a file (while writing to the file)
    %   #define     CONTINUE_GETFILE              0x97    //  Continue get byte from a file (while writing to the file)
    %   #define     CLOSE_FILEHANDLE              0x98    //  Close file handle
    %   #define     LIST_FILES                    0x99    //  List files
    %   #define     CONTINUE_LIST_FILES           0x9A    //  Continue list files
    %   #define     CREATE_DIR                    0x9B    //  Create directory
    %   #define     DELETE_FILE                   0x9C    //  Delete
    %   #define     LIST_OPEN_HANDLES             0x9D    //  List handles
    %   #define     WRITEMAILBOX                  0x9E    //  Write to mailbox
    %   #define     BLUETOOTHPIN                  0x9F    //  Transfer trusted pin code to brick
    %   #define     ENTERFWUPDATE                 0xA0    //  Restart the brick in Firmware update mode
    %   #define     SETBUNDLEID                   0xA1    //  Set Bundle ID for mode 2
    %   #define     SETBUNDLESEEDID               0xA2    //  Set bundle seed ID for mode 2
    % 
    % /*
    % 
    %   Byte 6 - n: Dependent on System Command
    % 
    % 
    % 
    %   System Command Response Bytes:
    %   ,------,------,------,------,------,------,------,------,
    %   |Byte 0|Byte 1|Byte 2|Byte 3|      |      |      |Byte n|
    %   '------'------'------'------'------'------'------'------'
    % 
    %   Byte 0 � 1: Reply size, Little Endian\n
    % 
    %   Byte 2 � 3: Message counter, Little Endian\n
    % 
    %   Byte 4:     Reply type. see following defines     */
    % 
    %   #define     SYSTEM_REPLY                  0x03    //  System command reply
    %   #define     SYSTEM_REPLY_ERROR            0x05    //  System command reply error
    % /*
    %   Byte 5:     System command this is the response to
    % 
    %   Byte 6:     Reply status
    % */
    % 
    %   // SYSTEM command return codes
    %   #define     SUCCESS                       0x00
    %   #define     UNKNOWN_HANDLE                0x01
    %   #define     HANDLE_NOT_READY              0x02
    %   #define     CORRUPT_FILE                  0x03
    %   #define     NO_HANDLES_AVAILABLE          0x04
    %   #define     NO_PERMISSION                 0x05
    %   #define     ILLEGAL_PATH                  0x06
    %   #define     FILE_EXITS                    0x07
    %   #define     END_OF_FILE                   0x08
    %   #define     SIZE_ERROR                    0x09
    %   #define     UNKNOWN_ERROR                 0x0A
    %   #define     ILLEGAL_FILENAME              0x0B
    %   #define     ILLEGAL_CONNECTION            0x0C
    % 
    % /*
    %   Byte 7 - n: Response dependent on System Command
    %
    %   Direct Command Bytes:
    %   ,------,------,------,------,------,------,------,------,
    %   |Byte 0|Byte 1|Byte 2|Byte 3|Byte 4|Byte 5|      |Byte n|
    %   '------'------'------'------'------'------'------'------'
    % 
    %   Byte 0 � 1: Command size, Little Endian\n
    % 
    %   Byte 2 � 3: Message counter, Little Endian\n
    % 
    %   Byte 4:     Command type. see following defines   */
    % 
    %   #define     DIRECT_COMMAND_REPLY          0x00    //  Direct command, reply required
    %   #define     DIRECT_COMMAND_NO_REPLY       0x80    //  Direct command, reply not required
    % 
    %                                                     /*
    % 
    %   Byte 5 - 6: Number of global and local variables (compressed).
    % 
    %                Byte 6    Byte 5
    %               76543210  76543210
    %               --------  --------
    %               llllllgg  gggggggg
    % 
    %                     gg  gggggggg  Global variables [0..MAX_COMMAND_GLOBALS]
    % 
    %               llllll              Local variables  [0..MAX_COMMAND_LOCALS]
    % 
    %   Byte 7 - n: Byte codes
    %
    %                     Direct Command Response Bytes:
    %   ,------,------,------,------,------,------,------,------,
    %   |Byte 0|Byte 1|Byte 2|Byte 3|      |      |      |Byte n|
    %   '------'------'------'------'------'------'------'------'
    % 
    %   Byte 0 - 1: Reply size, Little Endian\n
    % 
    %   Byte 2 - 3: Message counter, Little Endian\n
    % 
    %   Byte 4:     Reply type. see following defines     */
    % 
    %   #define     DIRECT_REPLY                  0x02    //  Direct command reply
    %   #define     DIRECT_REPLY_ERROR            0x04    //  Direct command reply error
    % 
    %                                                     /*
    % 
    %   Byte 5 - n: Response buffer (global variable values)
    %                                                     */
    
    properties
        msg
    end
    
    methods (Abstract)
        addHeader(data, varargin); 
        addLength(data);
    end
    
    methods
        function dataStream = DataStream(varargin)
            % DataStream.dataStream Create an empty DataStream
            %
            % c = DataStream() is an object that represents an EV3 DataStream
            % c = DataStream([...]) is an object that represents an EV3 reply
            %
            % Example::
            %           c = DataStream();
            
            if nargin > 0  % Packet is a reply
                try
                    dataStream.msg = uint8(varargin{1});
                catch ME
                    id = [ID(), ':', 'InvalidParameter'];
                    msg = 'Failed to create reply-packet with DataStream-class.';
                    baseException = MException(id, msg);
                    baseException = addCause(baseException, ME);
                    throw(baseException);
                end
                
                % If corrupt, throw error
                corrupt = dataStream.isValidReply();
                if corrupt < 1
                    msg = 'Invalid reply packet';
                    id = [ID(), ':', 'CorruptPacket'];
                    ME = MException(id, msg);
                    switch corrupt
                        case 0
                            causeMsg = 'Reply packet too short';
                        case -1
                            causeMsg = ['Specified packet length (bytes 1&2) does not equal ',...
                                        'actual packet length'];
                        case -2
                            causeMsg = 'Error type (byte 5) is not valid';
                    end
                    cause = MException(id, causeMsg);
                    ME = addCause(ME, cause);
                    throw(ME);
                end
            else  % Packet is a DataStream
                dataStream.msg = uint8([]);
            end
        end
        
        function errorDetected = checkForError(dataStream)
            % DataStream.checkForError Check DataStream for error
            %
            % error = checkForError() returns 1 if error byte in received
            % package is set and 0 if not.)
            % 
            % Note:: - If 5th byte is set to 0x04 (direct DataStream) or 0x05 
            %        (system DataStream) an error has occurred. 
            %        (found in EV3 Communication Developer Kit.pdf)
            %        - Only valid if dataStream is a reply packet, of course.
            
            % 5th byte in reply is used for error detection
            replyTypeOffset = 5; 
            errorDetected = 0;
            
            if dataStream.msg(replyTypeOffset) == 4 || dataStream.msg(replyTypeOffset) == 5
               errorDetected = 1; 
            end
        end
        
        function state = isValidReply(dataStream)
            % DataStream.isValidReply Check if reply-packet is valid
            %
            % state = dataStream.isValidReply checks dataStream against reply-packet-format and returns
            %                          whether if dataStream is valid (that is, does not fit the 
            %                          format).
            % Notes:
            %  - state = 1, if reply is not corrupted (at least not in a way that can be
            %               tested)
            %  - state = 0, if reply is too short (minimum length = 5 bytes)
            %  - state = -1, if reply is corrupted due to invalid length/length indicator
            %  - state = -2, if reply is corrupted due to invalid error type byte
            %  - IMPORTANT: Of course, the packet can still contain invalid information which
            %               cannot be tested properly without context. This corruption-test
            %               only checks the contextless information.
            %               
            state = 1;
            
            % Minimum length of a reply packet is 5, for a packet without response buffer
            minLength = 5;
            if length(dataStream.msg) < minLength
                state = 0; 
                return
            end
            
            % Note: The two things that can be checked without context are the length and the 
            %       DataStream type, the remainder varies from packet to packet. 
            pLength = double(typecast(dataStream.msg(1:2),'uint16')) + 2;
            pCmdType = dataStream.msg(5);
            
            if pLength ~= length(dataStream.msg)
                state = -1;
            elseif pCmdType ~= 2 && pCmdType ~=4
                state = -2;
            end
        end
        
        function addSystemDataStream(dataStream,v)
            % DataStream.addSystemDataStream Add a system DataStream
            %
            % DataStream.addSystemDataStream(v) adds a system DataStream to the
            % DataStream object.
            %
            % Notes::
            % - v is the system DataStreams which can be found in c_com.h
            % 
            % Example::
            %            dataStream.addSystemDataStream(SystemDataStreams.BeginDownload)
            
           dataStream.msg = [dataStream.msg uint8(v)]; 
        end
        
        function addDirectDataStream(dataStream,v)
            % DataStream.addDirectDataStream Add a direct DataStream
            %
            % DataStream.addDirectDataStream(v) adds a direct DataStream to the
            % DataStream object.
            %
            % Notes::
            % - v is the direct DataStreams opcode can be found in ByteCodes.m
            %
            % Example::
            %            dataStream.addDirectDataStream(SoundSubCodes.Tone)
            
           dataStream.msg = [dataStream.msg uint8(v)]; 
        end
        
        function clear(dataStream)
            % DataStream.clear Clear DataStream
            %
            % Commad.clear clears the message
            %
            % Example::
            %           dataStream.clear()

            dataStream.msg = '';
        end
        
        function s = char(dataStream)
            s = '';
            for i=1:length(dataStream.msg)
                s = [s sprintf(' %d', dataStream.msg(i))];
            end
        end
        
        function s = hex(dataStream)
            s = '';
            for i=1:length(dataStream.msg)
                s = [s sprintf(' %x', dataStream.msg(i))];
            end
        end
        
        function display(dataStream)
            % DataStream.display Display the DataStream message (decimal)
            %
            % DataStream.display() prints the DataStream message to the MATLAB
            % DataStream window in decimal format.
            %
            % Example::
            %           dataStream.display()
            
            loose = strcmp( get(0, 'FormatSpacing'), 'loose');
            if loose
                disp(' ');
            end
            disp([inputname(1), ' = '])
            disp( char(dataStream) );
        end 
        
        function displayHex(dataStream)
            % DataStream.displayHex Display the DataStream message (hex)
            %
            % DataStream.displayHex() prints the DataStream message to the MATLAB
            % DataStream window in hexadecimal format.
            %
            % Example::
            %           dataStream.displayHex()
            
            loose = strcmp( get(0, 'FormatSpacing'), 'loose');
            if loose
                disp(' ');
            end
            disp([inputname(1), ' = '])
            disp( hex(dataStream) );
        end 
        
        function LC0(dataStream,v)
            % DataStream.LC0 Add a local constant 0 
            %
            % DataStream.LC0(v) adds a local constant 0 to the DataStream object.
            %
            % Notes::
            % - v is the numerical value for the LC0 DataStream
            % - Local constant 0 is defined as ((v & PRIMPAR_VALUE) | PRIMPAR_SHORT |
            %   PRIMPAR_CONST) (which equals v)
            % - LC0 has a range from -32 to 31 since PRIMPAR_VALUE = 3F
            % 
            % Examples::
            %           dataStream.LC0(10) --> dataStream.msg = (dataStream.msg 10)
            
            dataStream.msg = [dataStream.msg bitor(bitand(typecast(int8(v),'uint8'),uint8(Primitives.pvalue)),bitor(uint8(Primitives.pshort),uint8(Primitives.pconst)))]; 
        end
        
        function LC1(dataStream,v)
            % DataStream.LC1 Add a local constant 1
            %
            % DataStream.LC1(v) adds a local constant 1 to the DataStream object.
            %
            % Notes::  
            % - LC1(v) is defined as (PRIMPAR_LONG  | PRIMPAR_CONST | PRIMPAR_1_BYTE),(v & 0xFF)
            % - LC1 has a range from -127 to 128    
            % - v is the numerical value for the LC1 DataStream
            % - resulting part of msg will be 2 bytes long:
            %   -> byte 1 = indicator(129) that 1 byte will follow 
            %   -> byte 2 = v
            % 
            % Examples::
            %           dataStream.LC1(100) --> dataStream.msg = (dataStream.msg 129 10)
            
            dataStream.msg = [dataStream.msg bitor(bitor(uint8(Primitives.plong),uint8(Primitives.pconst)),(uint8(Primitives.p1_byte))), bitand(typecast(int8(v),'uint8'),255)];
        end
        
        function LC2(dataStream,v)
            % DataStream.LC2 Add a local constant 2 
            %
            % DataStream.LC2(v) adds a local constant 2 to the DataStream object.
            %
            % Notes::
            % - Local constant 2 is defined as (PRIMPAR_LONG  | PRIMPAR_CONST | PRIMPAR_2_BYTES),(v & 0xFF),((v >> 8) & 0xFF)
            % - LC2 has a range from -32768 to 32767
            % - v is the numerical value for the LC2 DataStream
            % - resulting part of msg will be 3 bytes long:
            %   -> byte 1   = indicator(130) that 2 bytes will follow 
            %   -> byte 2-3 = v
            %
            % Examples::
            %           dataStream.LC2(1000) --> dataStream.msg = (dataStream.msg 130 232 3)
                        
            dataStream.msg = [dataStream.msg bitor(bitor(uint8(Primitives.plong),uint8(Primitives.pconst)),(uint8(Primitives.p2_byte))), typecast(int16(v), 'uint8')];
        end
        
        function LC4(dataStream,v)
            % DataStream.LC4 Add a local constant 4
            %
            % DataStream.LC4(v) adds a local constant 4 to the DataStream object.
            %
            % Notes::
            % - Local constant 4 is defined as(PRIMPAR_LONG  | PRIMPAR_CONST | PRIMPAR_4_BYTES),((ULONG)v & 0xFF),(((ULONG)v >> (ULONG)8) & 0xFF),(((ULONG)v >> (ULONG)16) & 0xFF),(((ULONG)v >> (ULONG)24) & 0xFF)
            % - LC4 has a range from -2,147,483,648 to 2,147,483,647
            % - v is the numerical value for the LC4 DataStream
            % - resulting part of msg will be 5 bytes long:
            %   -> byte 1   = indicator(131) that 4 bytes will follow 
            %   -> byte 2-5 = v 
            %
            % Examples::
            %           dataStream.LC4(10000) --> = (dataStream.msg 131 16 39 0 0)
            
            dataStream.msg = [dataStream.msg bitor(bitor(uint8(Primitives.plong),uint8(Primitives.pconst)),(uint8(Primitives.p4_byte))), typecast(int32(v), 'uint8')];
        end
        
        function LV0(dataStream,i)
            % DataStream.LV0 Add a local variable 0
            % 
            % DataStream.LV0(i) adds a local variable 0 to the DataStream object.
            %
            % Notes::
            % - Local variable 0 is defined as ((i & PRIMPAR_INDEX) | PRIMPAR_SHORT | PRIMPAR_VARIABEL | PRIMPAR_LOCAL)
            % - i is the number of 1 byte local variables needed.
            % 
            % Example::
            %           dataStream.LV0(1)
            
            dataStream.msg = [dataStream.msg bitor(bitor(bitor(bitand(uint8(i),Primitives.pindex),Primitives.pshort),Primitives.pvariabel),Primitives.plocal)];
        end
        
        function GV0(dataStream,i)
            % DataStream.GV0 Add a global variable 0
            %
            % DataStream.GV0(i) adds a global variable 0 to the DataStream
            % object.
            % 
            % Notes::
            % - Global variable 0 is defined as ((i & PRIMPAR_INDEX) | PRIMPAR_SHORT | PRIMPAR_VARIABEL | PRIMPAR_GLOBAL)
            % - i is the address in response buffer of reply packet where returned data should
            %   be stored
            % - the response buffer starts at byte 5 and ends at byte 5+nGV
            % - nGV defined in addHeaderDirectReply, used when creating DataStream in Brick.m
            % 
            % Example::
            %            dataStream.GV0(0)--> Starting at byte 5, first n (see addHeaderDirectReply) bytes 
            %                          in response buffer of reply packet are reserved for returning data
            %                          (n=1 for DATA8, n=4 for DATAF, etc.). 
            %                          Note: If nGV>n, remaining packet filled with either zeros or 
            %                                seemingly pointless data...
            % dataStream.GV0(0);dataStream.GV0(1)--> first byte in response buffer for DATA8, next nGV-1 bytes
            %                          for something else (e.g. DATA32, if nGV>=5)
            
            dataStream.msg = [dataStream.msg bitor(bitor(bitor(bitand(uint8(i),Primitives.pindex),Primitives.pshort),Primitives.pvariabel),Primitives.pglobal)];
        end
        
        % Implemented @ MMI
        % WIP
        function GV1(dataStream,i)
            % (PRIMPAR_LONG  | PRIMPAR_VARIABEL | PRIMPAR_GLOBAL | PRIMPAR_1_BYTE),(i & 0xFF)
            dataStream.msg = [dataStream.msg bitor(bitor(bitor(Primitives.pglobal,Primitives.p1_byte),Primitives.pvariabel),Primitives.plong) bitand(i,255)];
        end
        
        % Implemented @ MMI
        % WIP       
        function GV2(dataStream,i)
           % (PRIMPAR_LONG  | PRIMPAR_VARIABEL | PRIMPAR_GLOBAL | PRIMPAR_2_BYTES),(i & 0xFF),((i >> 8) & 0xFF) 
           dataStream.msg = [dataStream.msg bitor(bitor(bitor(Primitives.pglobal,Primitives.p2_byte),Primitives.pvariabel),Primitives.plong) bitand(i,255) bitand(bitshift(i,-8),255)];
        end
        
        % Implemented @ MMI
        % WIP
        function GV4(dataStream,i)
            % (PRIMPAR_LONG  | PRIMPAR_VARIABEL | PRIMPAR_GLOBAL | PRIMPAR_4_BYTES),(i & 0xFF),((i >> 8) & 0xFF),((i >> 16) & 0xFF),((i >> 24) & 0xFF)
            dataStream.msg = [dataStream.msg bitor(bitor(bitor(Primitives.pglobal,Primitives.p4_byte),Primitives.pvariabel),Primitives.plong) bitand(i,255) bitand(bitshift(i,-8),255) bitand(bitshift(i,-16),255) bitand(bitshift(i,-24),255)];
        end    
        
        function LCS(dataStream)
            % DataStream.LCS Add a local constant string
            %
            % DataStream.LCS() adds a local constant string to the DataStream
            % object.
            %
            % Notes::
            % - Local constant string is defined as (PRIMPAR_LONG | PRIMPAR_STRING)
            %
            % Example::
            %           dataStream.LCS()
            
            dataStream.msg = [dataStream.msg bitor(Primitives.plong,Primitives.pstring)];
        end
        
        function addByteCode(dataStream,v)
            % DataStream.addValue add a numerical value
            %
            % DataStream.addValue adds a numerical value to the DataStream
            % object.
            %
            % Notes::
            % - v is the numerical value to be added with range -128 to 127
            %
            % Example::
            %           dataStream.addValue(10)
            
            dataStream.msg = [dataStream.msg uint8(v)];
        end
        
        function addArray(dataStream,txt)
            % DataStream.addArray add a numerical array
            %
            % DataStream.addArray(txt) adds a numerical array to the DataStream
            % object.
            %
            % Notes::
            % - txt is the numerical array to be added
            % 
            % Example::
            %           dataStream.addArray([1,2,3,4,5])
            
            for i=1:length(txt)
               dataStream.addValue(txt(i)); 
            end   
        end
        
        function addString(dataStream,txt)
            % DataStream.addString add a string
            %
            % Commad.addString(txt) adds a string to the DataStream object.
            %
            % Notes::
            % - txt is the string to be added
            % - a null termination character is added to the end of the
            % string
            % - on most occasions addLCSString should be used
            %
            % Example::
            %           dataStream.addString('hello')
            
            for i=1:length(txt)
               dataStream.addValue(txt(i)); 
            end
            dataStream.addValue(0);    
        end
        
        function addLCSString(dataStream,txt)
            % DataStream.addLCSString add a string with the LCS type
            %
            % DataStream.addLCSString(txt) adds a string to the DataStream object
            % with the LCS DataStream as well.
            %
            % Notes::
            % - txt is the string to be added
            % - a null termination character is added to the end of the
            % string.
            %
            % Example::
            %           dataStream.addLCSString('hello')
            
            dataStream.LCS;
            for i=1:length(txt)
               dataStream.addValue(txt(i)); 
            end
            dataStream.addValue(0);    
        end
        
        function LONGToBytes(dataStream,x)
            % DataStream.LONGToBytes add a LONGToBytes
            %
            % DataStream.LONGToBytes(x) adds a LONGToBytes to the DataStream
            % object.
            %
            % Notes::
            % - x is the value to be added
            % - LONGToBytes is defined as(UBYTE)((_x) & 0xFF),(UBYTE)((_x >> 8) & 0xFF),(UBYTE)((_x >> 16) & 0xFF),(UBYTE)((_x >> 24) & 0xFF)
            %
            % Example::
            %           dataStream.LONGToBytes(1)
            
            dataStream.msg = [dataStream.msg typecast(uint32(x), 'uint8')];
        end
        
        function WORDToBytes(dataStream,x)
            % DataStream.WORDToBytes add a WORDToBytes
            %
            % DataStream.WORDToBytes(x) adds a WORDToBytes to the DataStream
            % object.
            %
            % Notes::
            % - x is the value to be added
            % - WORDToBytes is defined as (UBYTE)((_x) & 0xFF),(UBYTE)((_x >> 8) & 0xFF)
            %
            % Example::
            %           dataStream.WORDToBytes(1)
            
            dataStream.msg = [dataStream.msg typecast(uint16(x), 'uint8')];
        end
        
        function BYTEToBytes(dataStream,x)
            % DataStream.BYTEToBytes add a BYTEToBytes
            %
            % DataStream.BYTEToBytes(x) adds a BYTEToBytes to the DataStream
            % object.
            %
            % Notes::
            % - x is the value to be added
            % - BYTEToBytes is defined as (UBYTE)((_x) & 0xFF)
            %
            % Example::
            %           dataStream.BYTEToBytes(1)
            
            dataStream.msg = [dataStream.msg uint8(x)];
        end
        
        function GenerateByteCode(dataStream, fileName)
            % DataStream.GenerateByteCode Generate byte code
            %
            % DataStream.GenerateByteCode(fileName) prints the byte code in
            % the DataStream object to a file. This file can then be uplodaded
            % and executed directly on the brick.
            % 
            % Example::
            %           dataStream.GenerateByteCode('tst.rbf')

            fid = fopen([fileName '.rbf'], 'w');
            fwrite(fid,dataStream.msg,'uint8');
            fclose(fid);
            fprintf('Wrote %d bytes to %s.rbf\n',length(dataStream.msg),fileName);
        end
        
        function opUI_FLUSH(dataStream)
            % DataStream.opUI_FLUSH Add a opUI_FLUSH
            %
            % DataStream.opUI_FLUSH() adds a opUI_FLUSH opcode to the DataStream
            % object.
            %
            % Notes::
            % opUI_FLUSH
            %
            % Example::
            %           dataStream.opUI_FLUSH()
            
            dataStream.addDirectDataStream(ByteCodes.UIFlush);
        end
        
        function opUI_READ_GET_VBATT(dataStream,value)
            % DataStream.opUI_READ_GET_VBATT Add a opUI_READ_GET_VBATT
            %
            % DataStream.opUI_READ_GET_VBATT adds a opUI opcode with a 
            % READ_GET_VBATT subcode to the DataStream object.
            %
            % Notes::
            % - value is the address in response buffer where value (DATA32) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opUI_READ,LC0(1),GV0(value)
            %
            % Example::
            %           dataStream.opUI_READ_GET_VBATT(0)
            
            dataStream.addDirectDataStream(ByteCodes.UIRead);
            dataStream.LC0(UIReadSubCodes.GetVbatt);
            dataStream.GV0(value);
        end
        
        function opUI_READ_GET_LBATT(dataStream,value)
            % DataStream.opUI_READ_GET_LBATT Add a opUI_READ_GET_LBATT
            %
            % DataStream.opUI_READ_GET_LBATT adds a opUI opcode with a
            % READ_GET_LBATT subcode to the DataStream object.
            %
            % Notes::
            % - value is the address in response buffer where value (DATA8) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opUI_READ,LC0(18),GV0(value)
            %
            % Example::
            %           dataStream.opUI_READ_GET_LBATT(0)

            dataStream.addDirectDataStream(ByteCodes.UIRead);
            dataStream.LC0(UIReadSubCodes.GetLbatt);
            dataStream.GV0(value);
        end
        
        function opUI_WRITE_PUT_STRING(dataStream,txt)
            % DataStream.opUI_WRITE_PUT_STRING Add a opUI_WRITE_PUT_STRING
            %
            % DataStream.opUI_WRITE_PUT_STRING adds a opUR_WRITE opcode with a
            % WRITE_PUT_STRING subcode to the DataStream object.
            %
            % Notes::
            % - txt is the string to be added
            % - opUI_WRITE,LC0(8),LCS,'A','B' ... ,0
            %
            % Example::
            %           dataStream.opUI_WRITE_PUT_STRING('hello')
            
            dataStream.addDirectDataStream(ByteCodes.UIWrite);
            dataStream.LC0(UIWriteSubCodes.PutString);
            dataStream.addLCSString(txt);
        end
        
        function opUI_WRITE_INIT_RUN(dataStream)
            % DataStream.opUI_WRITE_INIT_RUN Add a opUI_WRITE_INIT_RUN
            %
            % DataStream.opUI_WRITE_INIT_RUN adds a opUI_WRITE opcode with a
            % INIT_RUN subcode to the DataStream object.
            %
            % Notes::
            % - opUI_WRITE,LC0(25)
            %
            % Example::
            %           dataStream.opUI_WRITE_INIT_RUN()
            
            dataStream.addDirectDataStream(ByteCodes.UIWrite);
            dataStream.LC0(UIWriteSubCodes.InitRun);
        end
        
        function opUI_WRITE_LED(dataStream,pattern)
            % DataStream.opUI_WRITE_LED Add a opUI_WRITE_LED
            %
            % DataStream.opUI_WRITE_LED(pattern) adds a opUI_WRITE opcode with
            % a WRITE_LED subcode to the DataStream object.
            %
            % Notes::
            % - pattern is the LED pattern from Device.LedBlack to
            % Device.LedOrangePulse
            % - opUI_WRITE,LC0(27),LC0(pattern)
            %
            % Example::
            %           dataStream.opUI_WRITE_LED(Device.LedBlack)
            
            dataStream.addDirectDataStream(ByteCodes.UIWrite);
            dataStream.LC0(UIWriteSubCodes.Led);
            dataStream.LC0(pattern);            
        end
        
        function opUI_DRAW_UPDATE(dataStream)
            % DataStream.opUI_DRAW_UPDATE Add a opUI_DRAW_UPDATE
            %
            % DataStream.opUI_DRAW_UPDATE() adds a opUI_DRAW opcode with a
            % DRAW_UPDATE subcode to the DataStream object.
            %
            % Notes::
            % - opUI_DRAW,LC0(0)
            %
            % Examples::
            %           dataStream.opUI_DRAW_UPDATE()
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Update);
        end
        
        function opUI_DRAW_CLEAN(dataStream)
            % DataStream.opUI_DRAW_CLEAN Add a opUI_DRAW_CLEAN
            %
            % DataStream.opUI_DRAW_CLEAN() adds a UI_DRAW opcode with a CLEAN
            % subcode to the DataStream object.
            %
            % Notes::
            % - opUI_DRAW,LC0(1)
            % 
            % Example::
            %           dataStream.opUI_DRAW_CLEAN()
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Clean);
        end
        
        function opUI_DRAW_PIXEL(dataStream,color,x,y)
            % DataStream.opUI_DRAW_PIXEL Add a opUI_DRAW_PIXEL
            %
            % DataStream.opUI_DRAW_PIXEL(color,x,y) adds a UI_DRAW opcode with
            % a PIXEL subcode to the DataStream object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x is the x coordinate on the LCD
            % - y is the y coordinate on the LCD
            % - opUI_DRAW,LC0(2),LC0(color),LC2(x),LC2(y)
            %
            % Example::
            %           dataStream.opUI_DRAW_PIXEL(vmCodes.vmFGColor,0,0)

            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Pixel);
            dataStream.LC0(color);
            dataStream.LC2(x);
            dataStream.LC2(y)
        end
        
        function opUI_DRAW_LINE(dataStream,color,x0,y0,x1,y1)
            % DataStream.opUI_DRAW_LINE Add a opUI_DRAW_LINE
            %
            % DataStream.opUI_DRAW_LINE(color,x0,y0,x1,y1) adds a UI_DRAW
            % opcode with a LINE subcode to the DataStream object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x0,y0 is the first point coordinate of the line
            % - x1,y1 is the second point coordinate of the line
            % - opUI_DRAW,LC0(3),LC0(color),LC2(x0),LC2(y0),LC2(x1),LC2(y1)
            %
            % Example::
            %           dataStream.opUI_DRAW_LINE(vmCodes.vmFGColor,0,0,1,1)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Line);
            dataStream.LC0(color);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(x1);
            dataStream.LC2(y1);
        end
        
        function opUI_DRAW_CIRCLE(dataStream,color,x0,y0,r)
            % DataStream.opUI_DRAW_CIRCLE Add a opUI_DRAW_CIRCLE
            %
            % DataStream.opUI_DRAW_CIRCLE(color,x0,y0,r) adds a UI_DRAW opcode
            % with a CIRCLE subcode to the object DataStream.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x is the x coordinate of the circle center
            % - y is the y coordinate of the circle center
            % - r is the radius of the circle 
            % - opUI_DRAW,LC0(4),LC0(color),LC2(x0),LC2(y0),LC2(r)
            %
            % Example::
            %           dataStream.opUI_DRAW_CIRCLE(vmCodes.vmFGColor,0,0,5)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Circle);
            dataStream.LC0(color);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(r);
        end           
        
        function opUI_DRAW_TEXT(dataStream,color,x,y,txt)
            % DataStream.opUI_DRAW_TEXT Add a opUI_DRAW_TEXT
            %
            % DataStream.opUI_DRAW_TEXT(color,x,y,txt) adds a UI_DRAW opcode
            % with a TEXT subcode to the DataStream object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x is the x start coordinate of the text
            % - y is the y start coordinate of the text
            % - txt is the text to be printed
            % - opUI_DRAW,LC0(5),LC0(color),LC2(x),LC2(y),LCS,'A','B' ...,0
            %
            % Example::
            %           dataStream.opUI_DRAW_TEXT(vmCodes.vmFGColor,0,0,'hello')
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Text);
            dataStream.LC0(color);
            dataStream.LC2(x);
            dataStream.LC2(y);
            dataStream.addLCSString(txt);
        end
        
        function opUI_DRAW_VALUE(dataStream,color,x,y,index,figures,decimals)
            % DataStream.opUI_DRAW_VALUE Add a opUI_DRAW_VALUE
            %
            % DataStream.opUI_DRAW_VALUE(color,x,y,index,figures,decimals)
            % adds a UI_DRAW opcode with a VALUE subcode to the DataStream
            % object.
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x is the x start coordinate of the text
            % - y is the y start coordinate of the text
            % - index is the value to write at an index
            % - figures is the total number of figured inclusive decimal point
            % - decimals is the number of decimals
            % - opUI_DRAW,LC0(8),LC0(color),LC2(x),LC2(y),GV0(index),LC0(figures),LC0(decimals)
            %
            % Example::
            %           dataStream.opUI_DRAW_VALUE(vmCodes.vmFGColor,0,0,0,3,2)
            %
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Value);
            dataStream.LC0(color);
            dataStream.LC2(x);
            dataStream.LC2(y);
            dataStream.GV0(index);
            dataStream.LC0(figures);
            dataStream.LC0(decimals);
        end
        
        function opUI_DRAW_FILLRECT(dataStream,color,x0,y0,x1,y1)
            % DataStream.opUI_DRAW_FILLRECT Add a opUI_DRAW_FILLRECT
            %
            % DataStream.opUI_DRAW_FILLRECT(color,x0,y0,x1,y1) adds a
            % opUI_DRAW opcode with a FILLRECT subcode to the DataStream
            % object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x0,y0 is the point coordinate of the top left corner
            % - x1,y1 is the point cooredinate of the bottom right corner
            % - opUI_DRAW,LC0(9),LC0(color),LC2(x0),LC2(y0),LC2(x1),LC2(y1)
            %
            % Example::
            %           dataStream.opUI_DRAW_FILLRECT(vmCodes.vmFGColor,0,0,10,10)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Fillrect);
            dataStream.LC0(color);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(x1);
            dataStream.LC2(y1);
        end
        
        function opUI_DRAW_RECT(dataStream,color,x0,y0,x1,y1)
            % DataStream.opUI_DRAW_RECT Add a opUI_DRAW_RECT
            %
            % DataStream.opUI_DRAW_RECT(color,x0,y0,x1,y1) adds a
            % opUI_DRAW opcode with a RECT subcode to the DataStream
            % object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - x0,y0 is the point coordinate of the top left corner
            % - x1,y1 is the point cooredinate of the bottom right corner
            % - opUI_DRAW,LC0(10),LC0(color),LC2(x0),LC2(y0),LC2(x1),LC2(y1)
            %
            % Example::
            %           dataStream.opUI_DRAW_RECT(vmCodes.vmFGColor,0,0,10,10)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Rect);
            dataStream.LC0(color);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(x1);
            dataStream.LC2(y1);
        end
        
        function opUI_DRAW_INVERSERECT(dataStream,x0,y0,x1,y1)
            % DataStream.opUI_DRAW_INVERSERECT Add a opUI_DRAW_INVERSERECT
            %
            % DataStream.opUI_DRAW_INVERSERECT(x0,y0,x1,y1) adds a opUI_DRAW
            % opcode with a INVERSERECT subcode to the DataStream object.
            %
            % Notes::
            % - x0,y0 is the point coordinate of the top left corner
            % - x1,y1 is the point cooredinate of the bottom right corner
            % - opUI_DRAW,LC0(16),LC2(x0),LC2(y0),LC2(x1),LC2(y1)
            %
            % Examples::
            %           dataStream.opUI_DRAW_INVERSERECT(0,0,10,10)

            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Inverserect);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(x1);
            dataStream.LC2(y1);
        end
        
        function opUI_DRAW_SELECT_FONT(dataStream,type)
            % DataStream.opUI_DRAW_SELECT_FONT Add a opUI_DRAW_SELECT_FONT
            %
            % DataStream.opUI_DRAW_SELECT_FONT(type) adds a opUI_DRAW opcode
            % with a SELECT_FONT subcode to the DataStream object.
            %
            % Notes::
            % - type is the font type [0..2] 
            % - opUI_DRAW,LC0(17),LC0(type)
            %
            % Examples::
            %           dataStream.opUI_DRAW_SELECT_FONT(1)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.SelectFont);
            dataStream.LC0(type);
        end
        
        function opUI_DRAW_TOPLINE(dataStream,enable)
            % DataStream.opUI_DRAW_TOPLINE Add a opUI_DRAW_TOPLINE
            %
            % DataStream.opUI_DRAW_TOPLINE(enable) adds a opUI_DRAW opcode
            % with a TOPLINE subcode to the DataStream object.
            %
            % Notes::
            % - enable is the top status line flag, 0 = disable, 1 = enable
            % - opUI_DRAW,LC0(18),LC0(enable)
            %
            % Example::
            %           dataStream.opUI_DRAW_TOPLINE(1)
         
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Topline);
            dataStream.LC0(enable);
        end
        
        function opUI_DRAW_FILLWINDOW(dataStream, color, y0, y1)
            % DataStream.opUI_DRAW_FILLWINDOW Add a opUI_DRAW_FILLWINDOW
            %
            % DataStream.opUI_DRAW_FILLWINDOW adds a UI_DRAW opcode with a
            % FILLWINDOW subcode to the DataStream object.
            %
            % Notes::
            % - color is the pixel color (either foregrond or background)
            % - y0 is the start Y coordinate
            % - y1 is the size of Y to fill to
            % - opUI_DRAW,LC0(19),LC0(color),LC2(y0),LC2(y1)
            %
            % Example::
            %           dataStream.opUI_DRAW_FILLWINDOW(vmCodes.vmFGColor,0,10)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Fillwindow);
            dataStream.LC0(color);
            dataStream.LC2(y0);
            dataStream.LC2(y1);            
        end
        
        function opUI_DRAW_FILLCIRCLE(dataStream,color,x0,y0,r)
            % DataStream.opUI_DRAW_FILLCIRCLE Add a opUI_DRAW_FILLCIRCLE
            %
            % DataStream.opUI_DRAW_FILLCIRCLE(color,x0,y0,r) adds a UI_DRAW
            % opcode with a FILLCIRCLE subcode to the DataStream object.
            %
            % Notes::
            % - color is the pixel color (either foreground or background)
            % - x0 is the x coordinate of the circle center
            % - y0 is the y coordinate of the circl center
            % - r is the radius of the circle
            % - opUI_DRAW,LC0(24),LC0(color),LC2(x0),LC2(y0),LC2(r)
            %
            % Example::
            %           dataStream.opUI_DRAW_FILLCIRCLE(vmCodes.vmFGColor,10,10,5)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Fillcircle);
            dataStream.LC0(color);
            dataStream.LC2(x0);
            dataStream.LC2(y0);
            dataStream.LC2(r);
        end
        
        function opUI_DRAW_STORE(dataStream,no)
            % DataStream.opUI_DRAW_STORE Add a opUI_DRAW_STORE
            %
            % DataStream.opUI_DRAW_STORE(no) adds a UI_DRAW opcode with a
            % STORE subcode to the DataStream object. 
            %
            % Notes::
            % - no is the level number to store the UI screen
            % - opUI_DRAW,LC0(25),LC0(no)
            %
            % Example::
            %           dataStream.opUI_DRAW_STORE(1)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Store);
            dataStream.LC0(no);
        end
        
        function opUI_DRAW_RESTORE(dataStream,no)
            % DataStream.opUI_DRAW_RESTORE Add a opUI_DRAW_RESTORE
            %
            % DataStream.opUI_DRAW_RESTORE(no) adds a UI_DRAW opcode with a
            % RESTORE subcode to the DataStream object. 
            %
            % Notes::
            % - no is the level number to store the UI screen (0 is saved screen before run)
            % - opUI_DRAW,LC0(26),LC0(no)
            %
            % Example::
            %           dataStream.opUI_DRAW_RESTORE(1)
            
            dataStream.addDirectDataStream(ByteCodes.UIDraw);
            dataStream.LC0(UIDrawSubCodes.Restore);
            dataStream.LC0(no);
        end
        
        function opTIMER_WAIT(dataStream,time,timer)
            % DataStream.opTIMER_WAIT Add a opTIMER_WAIT
            %
            % DataStream.opTIMER_WAIT(time,timer) adds a opTIMER opcode with a
            % WAIT subcode to the DataStream object.
            %
            % Notes::
            % - time is the time is wait in ms
            % - timer is the local variable used for timing
            % - opTIMER_WAIT,LC2(time),LV0(timer)
            %
            % Example::
            %           dataStream.opTIMER_WAIT(1000,0)
            
            dataStream.addDirectDataStream(ByteCodes.TimerWait);
            dataStream.LC2(time);
            dataStream.LV0(timer);
        end
        
        function opTIMER_READY(dataStream,timer)
            % DataStream.opTIMER_READY Add a opTIMER_READY
            %
            % DataStream.opTIMER_READY(timer) adds a opTIMER opcode with a
            % READY subcode to the DataStream object.
            %
            % Notes::
            % - timer is the local variable used for timing
            % - opTIMER_READY,LV0(timer)
            %
            % Example::
            %           dataStream.opTIMER_READY(0)
            
            dataStream.addDirectDataStream(ByteCodes.TimerReady);
            dataStream.LV0(timer);
        end
        
        function opTIMER_READ(dataStream,time)
            % DataStream.opTIMER_READ Add a opTIMER_READ
            %
            % DataStream.opTIMER_READ(time) adds a opTIMER opcode with a READ
            % subcode to the DataStream object.
            %
            % Notes::
            % - time is the timer to be read in ms
            % - opTIMER_READ,LV0(time)
            %
            % Example::
            %           dataStream.opTIMER_READ(0)
            
            dataStream.addDirectDataStream(ByteCodes.TimerRead);
            dataStream.LV0(time);
        end
        
        function opSOUND_BREAK(dataStream)
            % DataStream.opSOUND_BREAK Add a opSOUND_BREAK
            %
            % DataStream.opSOUND_BREAK() adds a opSOUND opcode with a BREAK
            % subcode to the DataStream object.
            %
            % Notes::
            % - opSOUND,LC0(0)
            %
            % Example::
            %           dataStream.opSOUND_BREAK()
            
            dataStream.addDirectDataStream(ByteCodes.Sound);
            dataStream.LC0(SoundSubCodes.Break);
        end
        
        function opSOUND_TONE(dataStream,volume,frequency,duration)
            % DataStream.opSOUND_TONE Add a opSOUND_TONE
            %
            % DataStream.opSOUND_TONE(volume,frequency,duration) adds a
            % opSOUND opcode with a TONE subcode to the DataStream object.
            %
            % Notes::
            % - volume is the tone volume from 0 to 100
            % - frequency is the tone frequency in Hz
            % - duration is the tone duration in ms
            % - opSOUND,LC0(1),LC1(volume),LC2(frequency),LC2(duration)
            %
            % Example::
            %           dataStream.opSOUND_TONE(5,1000,500)
            
            dataStream.addDirectDataStream(ByteCodes.Sound);
            dataStream.LC0(SoundSubCodes.Tone);
            dataStream.LC1(volume);
            dataStream.LC2(frequency);
            dataStream.LC2(duration);
        end
        
        function opSOUND_PLAY(dataStream,volume,name)
            % DataStream.opSOUND_PLAY Add a opSOUND_PLAY
            %
            % DataStream.opSOUND_PLAY(volume,name) adds a opSOUND opcode with
            % a PLAY subcode the DataStream object.
            %
            % Example::
            %           dataStream.opSOUND_PLAY(5,'test.rsf')
            %
            % Notes::
            % - volume is the volume to play the file at
            % - name is the name of the file to play
            % - opSOUND,LC0(2),LC0(volume),LCS,'A','B' ... '0'
            
            dataStream.addDirectDataStream(ByteCodes.Sound);
            dataStream.LC0(SoundSubCodes.Play);
            dataStream.LC0(volume);
            dataStream.addLCSString(name);
        end
        
        function opSOUND_REPEAT(dataStream,volume,name)
            % DataStream.opSOUND_REPEAT Add a opSOUND_REPEAT
            %
            % DataStream.opSOUND_REPEAT(volume,name) adds a opSOUND opcode
            % with a REPEAT subcode to the DataStream object.
            %
            % Notes::
            % - volume is the volume to repeat the file at
            % - name is the name of the file to repeat
            % - opSOUND,LC0(2),LC0(volume),LCS,'A','B' ... '0'
            %
            % Example::
            %           dataStream.opSOUND_REPEAT(1,'test.rsf')
            
            dataStream.addDirectDataStream(ByteCodes.Sound);
            dataStream.addDirectDataStream(SoundSubCodes.Repeat);
            dataStream.LC0(volume);
            dataStream.addLCSString(name);
        end
        
        function opSOUND_TEST(dataStream,busy)
            % DataStream.opSOUND_TEST Add a opSOUND_TEST
            %
            % DataStream.opSOUND_TEST(busy) adds a opSOUND_TEST to the DataStream
            % object. 
            %
            % Notes::
            % - busy is the address in response buffer where the busy flag (DATA8) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opSOUND_TEST,GV0(busy)
            %
            % Example::
            %           dataStream.opSOUND_TEST(0)
            
            dataStream.addDirectDataStream(ByteCodes.SoundTest);
            dataStream.GV0(busy);
        end
        
        function opSOUND_READY(dataStream)
            % Commad.opSOUND_READY Add a opSOUND_READY
            %
            % DataStream.opSOUND_READY() adds a opSOUND_READY opcode to the 
            % DataStream object.
            %
            % Example::
            %           dataStream.opSOUND_READY()
            %
            % Notes::
            % - opSOUND_READY
            
            dataStream.addDirectDataStream(ByteCodes.SoundReady);           
        end
        
        function opINPUT_DEVICE_LIST(dataStream,length,array,changed)
            % DataStream.opINPUT_DEVICE_LIST Add a opINPUT_DEVICE_LIST
            %
            % DataStream.opINPUT_DEVICE_LIST(length,array,changed) adds a
            % opINPUT_DEVICE_LIST opcode to the DataStream object.
            %
            % Notes::
            % - length is maximum number of device types (norm 32)
            % - array is the returned device array list (DATA8 but can be 32 long so DATA32)
            % - changed is the returned device changed status (DATA8)
            % - opINPUT_DEVICE_LIST,LC0(length),GV0(array),GV0(changed)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_LIST(32,0,4)
            
            dataStream.addDirectDataStream(ByteCodes.InputDeviceList);
            dataStream.LC0(length);
            dataStream.GV0(array);
            dataStream.GV0(changed);
        end
        
        function opINPUT_DEVICE_GET_TYPEMODE(dataStream,layer,no,type,mode)
            % DataStream.opINPUT_DEVICE_GET_TYPEMODE Add a opINPUT_DEVICE_GET_TYPEMODE
            %
            % DataStream.opINPUT_DEVICE_GET_TYPEMODE(layer,no,type,mode) adds
            % a opINPUT_DEVICE opcode with a GET_TYPEMODE subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - type is the address in response buffer where type (DATA8) should be stored
            % - mode is the address in response buffer where mode (DATA8) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opINPUT_DEVICE,LC0(5),LC0(layer),LC0(no),GV0(type),GV0(mode)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_TYPEMODE(0,SensorPort.Sensor1,0,1)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetTypeMode);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(type);
            dataStream.GV0(mode);
        end
        
        function opINPUT_DEVICE_SET_TYPEMODE(dataStream,oldType,oldMode,newType,newMode)
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.SetTypeMode);
            dataStream.LC0(oldType);
            dataStream.LC0(oldMode);
            dataStream.LC0(newType);
            dataStream.LC0(newMode);
        end
        
        function opINPUT_DEVICE_GET_SYMBOL(dataStream,layer,no,length,destination)
            % DataStream.opINPUT_DEVICE_GET_SYMBOL Add a opINPUT_DEVICE_GET_SYMBOL
            %
            % DataStream.opINPUT_DEVICE_GET_SYMBOL(layer,no,length,destination)
            % adds a opINPUT_DEVICE opcode with a GET_SYMBOL subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - length is the maximal length of string returned (-1 no check)
            % - destination is the address in response buffer where string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opINPUT_DEVICE,LC0(6),LC0(layer),LC0(no),LC0(length),GV0(destination)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_SYMBOL(0,SensorPort.Sensor1,5,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetSymbol);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.LC0(length);
            dataStream.GV0(destination);
        end
        
        function opINPUT_DEVICE_CLR_ALL(dataStream,layer)
            % DataStream.opINPUT_DEVICE_CLR_ALL Add a opINPUT_DEVICE_CLR_ALL
            %
            % DataStream.opINPUT_DEVICE_CLR_ALL(layer) adds a opINPUT_DEVICE
            % opcode with a CLR_ALL subcode to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_CLR_ALL(0)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.ClrAll);
            dataStream.LC0(layer);
        end 
        
        % Implemented @ MMI
        function opINPUT_DEVICE_CLR_CHANGES(dataStream,layer,no)
            % DataStream.opINPUT_DEVICE_CLR_CHANGES Add a opINPUT_DEVICE_CLR_CHANGES
            %
            % DataStream.opINPUT_DEVICE_CLR_CHANGES(layer,no)
            % adds a opINPUT_DEVICE opcode with a CLR_CHANGES subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - opINPUT_DEVICE,LC0(26),LC0(layer),LC0(no)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_CLR_CHANGES(0,SensorPort.Sensor1)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.ClrChanges);
            dataStream.LC0(layer);
            dataStream.LC0(no);
        end        
        
        function opINPUT_DEVICE_GET_NAME(dataStream,layer,no,length,destination)
            % DataStream.opINPUT_DEVICE_GET_NAME Add a opINPUT_DEVICE_GET_NAME
            %
            % DataStream.opINPUT_DEVICE_GET_NAME(layer,no,length,destination)
            % adds a opINPUT_DEVICE opcode with a GET_NAME subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - length is the maximal length of string returned (-1 no check)
            % - destination is the address in response buffer where string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opINPUT_DEVICE,LC0(21),LC0(layer),LC0(no),LC0(length),GV0(destination)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_NAME(0,SensorPort.Sensor1,12,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetName);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.LC0(length);
            dataStream.GV0(destination);
        end 
        
        % Implemented @ MMI
        function opINPUT_DEVICE_GET_MODENAME(dataStream,layer,no,mode,length,destination)
            % DataStream.opINPUT_DEVICE_GET_MODENAME Add a opINPUT_DEVICE_GET_MODENAME
            %
            % DataStream.opINPUT_DEVICE_GET_MODENAME(layer,no,mode,length,destination)
            % adds a opINPUT_DEVICE opcode with a GET_MODENAME subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - mode is the sensor mode at NO
            % - length is the maximal length of string returned (-1 no check)
            % - destination is the address in response buffer where string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opINPUT_DEVICE,LC0(22),LC0(layer),LC0(no),LC0(mode),LC0(length),GV0(destination)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_MODENAME(0,SensorPort.Sensor1,Device.Bumps,12,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetModeName);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.LC0(mode);
            dataStream.LC0(length);
            dataStream.GV0(destination);
        end 

        % Implemented @ MMI
        function opINPUT_DEVICE_GET_CONNECTION(dataStream,layer,no,connection)
            % DataStream.opINPUT_DEVICE_GET_CONNECTION Add a
            % opINPUT_DEVICE_GET_CONNECTION
            %
            % DataStream.opINPUT_DEVICE_GET_CONNECTION(layer,no,connection)
            % adds a opINPUT_DEVICE opcode with a GET_CONNECTION subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - connection is the address in response buffer where the connection type (DATA8) 
            %   should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)            
            % - opINPUT_DEVICE,LC0(12),LC0(layer),LC0(no),GV0(connection)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_CONNECTION(0,SensorPort.Sensor1,0)  
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetConnection);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(connection);
        end
        
        % Implemented @ MMI
        function opINPUT_DEVICE_GET_MINMAX(dataStream,layer,no,min,max)
            % DataStream.opINPUT_DEVICE_GET_MINMAX Add a
            % opINPUT_DEVICE_GET_MINMAX
            %
            % DataStream.opINPUT_DEVICE_GET_MINMAX(layer,no,min,max)
            % adds a opINPUT_DEVICE opcode with a GET_MINMAX subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - min is the address in response buffer where the min val (DATAF) should be stored
            % - max is the address in response buffer where the max val (DATAF) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opINPUT_DEVICE,LC0(30),LC0(layer),LC0(no),GV0(min),GV0(max)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_MINMAX(0,SensorPort.Sensor1,0,4)  
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetMinMax);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(min);
            dataStream.GV0(max);
%             dataStream.GV4(min);
%             dataStream.GV4(max);
        end

        % Implemented @ MMI
        function opINPUT_DEVICE_GET_CHANGES(dataStream,layer,no,changes)
            % DataStream.opINPUT_DEVICE_GET_CHANGES Add a
            % opINPUT_DEVICE_GET_CHANGES
            %
            % DataStream.opINPUT_DEVICE_GET_CHANGES(layer,no,min,max)
            % adds a opINPUT_DEVICE opcode with a GET_CHANGES subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - changes is the address in response buffer where changes (DATAF) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opINPUT_DEVICE,LC0(25),LC0(layer),LC0(no),GV0(changes)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_CHANGES(0,SensorPort.Sensor1,0)  
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetChanges);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(changes);
        end
        
        % Implemented @ MMI
        function opINPUT_DEVICE_GET_FORMAT(dataStream,layer,no,datasets,format,modes,view)
            % DataStream.opINPUT_DEVICE_GET_FORMAT Add a
            % opINPUT_DEVICE_GET_FORMAT
            %
            % DataStream.opINPUT_DEVICE_GET_FORMAT(layer,no,min,max)
            % adds a opINPUT_DEVICE opcode with a GET_FORMAT subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - datasets is the address in response buffer where the no of datasets (DATA8) 
            %   should be stored
            % - format is the address in response buffer where the format (DATA8) should be 
            %   stored
            % - modes is the address in response buffer where the no of modes (DATA8) should be
            %   stored
            % - view is the address in response buffer where the view (DATA8) should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % opINPUT_DEVICE,LC0(2),LC0(layer),LC0(no),GV0(datasets),
            %                       GV0(format),GV0(modes),GV0(view)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_FORMAT(0,SensorPort.Sensor1,0,1,2,3)              
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetFormat);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(datasets);
            dataStream.GV0(format);
            dataStream.GV0(modes);
            dataStream.GV0(view);
        end
        
        % Implemented @ MMI
        function opINPUT_DEVICE_GET_BUMPS(dataStream,layer,no,bumps)
            % DataStream.opINPUT_DEVICE_GET_BUMPS Add a
            % opINPUT_DEVICE_GET_BUMPS
            %
            % DataStream.opINPUT_DEVICE_GET_BUMPS(layer,no,bumps)
            % adds a opINPUT_DEVICE opcode with a GET_BUMPS subcode to the
            % DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - bumps is the address in response buffer where the no of bumps (DATA8)should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            %
            % opINPUT_DEVICE,LC0(31),LC0(layer),LC0(no),GV0(bumps)
            %
            % Example::
            %           dataStream.opINPUT_DEVICE_GET_BUMPS(0,SensorPort.Sensor1,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputDevice);
            dataStream.LC0(InputDeviceSubCodes.GetBumps);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(bumps);
        end
                
        function opINPUT_READY(dataStream,layer,no)
            % DataStream.opINPUT_READY Add a opINPUT_READY
            %
            % DataStream.opINPUT_READY(layer,no) adds a opINPUT_READY opcode
            % to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - opINPUT_READY,LC0(layer),LC0(no)
            %
            % Example::
            %           dataStream.opINPUT_READY(0,SensorPort.Sensor1)
            
            dataStream.addDirectDataStream(ByteCodes.InputReady);
            dataStream.LC0(layer);
            dataStream.LC0(no);
        end
        
        % Implemented @ MMI
        function opINPUT_TEST(dataStream,layer,no,busy)
            % DataStream.opINPUT_TEST Add a opINPUT_TEST
            %
            % DataStream.opINPUT_TEST(layer,no,value) adds a opINPUT_TEST opcode
            % to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - busy is the address in response buffer where the busy flag (DATA8) should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opINPUT_TEST,LC0(layer),LC0(no),GV0(value)
            %
            % Example::
            %           dataStream.opINPUT_TEST(0,SensorPort.Sensor1,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputTest);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(value);
        end
        
        function opINPUT_READ(dataStream,layer,no,type,mode,pct)
            % DataStream.opINPUT_READ Add a opINPUT_READ
            %
            % DataStream.opINPUT_READ(layer,no,type,mode,pct) adds a opINPUT_READ 
            % opcode to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - type is the sensor type from types.html (0 = don't change)
            % - mode is the sensor mode from types.html [0..7] (-1 = don't change)
            % - pct is the address in response buffer where the read pct (DATA8) 
            %   should be stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - The device type is already set when the sensor is
            % connected, you should only need to specify the mode
            % - opINPUT_READ,LC0(layer),LC0(no),LC0(type),LC0(mode),GV0(pct)
            %
            % Example::
            %           dataStream.opINPUT_READ(0,SensorPort.Sensor1,Device.Ultrasonic,Device.USDistCM,0)
            
            dataStream.addDirectDataStream(ByteCodes.InputRead);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.LC0(type);
            dataStream.LC0(mode);
            dataStream.GV0(pct);
        end
        
        function opINPUT_READSI(dataStream,layer,no,type,mode,si)
            % DataStream.opINPUT_READSI Add a opINPUT_READSI
            %
            % DataStream.opINPUT_READSI(layer,no,type,mode,si) adds a opINPUT_READSI  
            % opcode to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is the output port number from [0..3] or sensor port number minus 1
            % - type is the sensor type from types.html (0 = don't change)
            % - mode is the sensor mode from types.html [0..7] (-1 = don't change)
            % - si is the address in response buffer where the read si value (DATA32) should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - The device type is already set when the sensor is connected, you should only 
            %   need to specify the mode
            % - opINPUT_READSI,LC0(layer),LC0(no),LC0(type),LC0(mode),GV0(si)
            %
            % Example::
            %           dataStream.opINPUT_READ(0,SensorPort.Sensor1,Device.Ultrasonic,Device.USDistCM,0)

            dataStream.addDirectDataStream(ByteCodes.InputReadSI);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.LC0(type);
            dataStream.LC0(mode);
            dataStream.GV0(si);
        end
        
        function opOUTPUT_SET_TYPE(dataStream,layer,nos,type)
            % DataStream.opOUTPUT_SET_TYPE Add a opOUTPUT_SET_TYPE
            %
            % DataStream.opOUTPUT_SET_TYPE(layer,nos,type) adds a opOUTPUT_SET_TYPE
            % opcode to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - type is the output device type from 0 to 255 (see types.html)
            % - opOUTPUT_SET_TYPE,LC0(layer),LC0(nos),LC0(type)
            %
            % Example::
            %           dataStream.opOUTPUT_SET_TYPE(0,MotorBitfield.MotorA,0)
            
            dataStream.addDirectDataStream(ByteCodes.OutputSetType);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC0(type);
        end
        
        function opOUTPUT_RESET(dataStream,layer,nos)
            % DataStream.opOUTPUT_RESET Add a opOUTPUT_RESET
            %
            % DataStream.opOUTPUT_RESET(layer,nos) adds a opOUTPUT_RESET opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - opOUTPUT_RESET,LC0(layer),LC0(nos)
            %
            % Example::
            %           dataStream.opOUTPUT_RESET(0,MotorBitfield.MotorA)
                        
            dataStream.addDirectDataStream(ByteCodes.OutputReset);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
        end
        
        function opOUTPUT_STOP(dataStream,layer,nos,brake)
            % DataStream.opOUTPUT_STOP Add a opOUTPUT_STOP
            %
            % DataStream.opOUTPUT_STOP(layer,nos,brake) adds a opOUTPUT_STOP opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - opOUTPUT_STOP,LC0(layer),LC0(nos),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_STOP(0,MotorBitfield.MotorA,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputStop);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC0(brake);
        end
        
        function opOUTPUT_SPEED(dataStream,layer,nos,speed)
            % DataStream.opOUTPUT_SPEED Add a opOUTPUT_SPEED
            %
            % DataStream.opOUTPUT_SPEED(layer,nos,speed) adds a opOUTPUT_SPEED opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - speed is the output speed with [+-0..100%] range
            % - opOUTPUT_POWER,LC0(layer),LC0(nos),LC1(speed)
            %
            % Example::
            %           dataStream.opOUTPUT_SPEED(0,MotorBitfield.MotorA,10)
           
            dataStream.addDirectDataStream(ByteCodes.OutputSpeed);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(speed);
        end
        
        function opOUTPUT_POWER(dataStream,layer,nos,power)
            % DataStream.opOUTPUT_POWER Add a opOUTPUT_POWER
            %
            % DataStream.opOUTPUT_POWER(layer,nos,speed) adds a opOUTPUT_POWER opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - power is the output power with [+-0..100%] range
            % - opOUTPUT_POWER,LC0(layer),LC0(nos),LC1(power)
            %
            % Example::
            %           dataStream.opOUTPUT_POWER(0,MotorBitfield.MotorA,10)
            
            dataStream.addDirectDataStream(ByteCodes.OutputPower);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(power);
        end
        
        function opOUTPUT_START(dataStream,layer,nos)
            % DataStream.opOUTPUT_START Add a opOUTPUT_START
            %
            % DataStream.opOUTPUT_START(layer,nos) adds a opOUTPUT_START opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - opOUTPUT_START,LC0(layer),LC0(nos)
            %
            % Example::
            %           dataStream.opOUTPUT_START(0,MotorBitfield.MotorA)
            
            dataStream.addByteCode(ByteCodes.OutputStart);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
        end
        
        function opOUTPUT_POLARITY(dataStream,layer,nos,pol)
            % DataStream.opOUTPUT_POLARITY Add a opOUTPUT_POLARITY
            %
            % DataStream.opOUTPUT_POLARITY(layer,nos,pol) adds a opOUTPUT_POLARITY opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - pol is the polarity [-1,0,1], -1 makes the motor run
            % backward, 1 makes the motor run forward, 0 makes the motor
            % run the opposite direction
            % - opOUTPUT_POLARITY,LC0(layer),LC0(nos),LC0(pol)
            %
            % Example::
            %           dataStream.opOUTPUT_POLARITY(0,MotorBitfield.MotorA,0)
            
            dataStream.addDirectDataStream(ByteCodes.OutputPolarity);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC0(pol);
        end
        
        function opOUTPUT_READ(dataStream,layer,no,speed,tacho)
            % DataStream.opOUTPUT_READ Add a opOUTPUT_READ
            %
            % DataStream.opOUTPUT_READ(layer,no,speed,tacho) adds a opOUTPUT_READ opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - no is output motor number [0...3] ( != NOS)
            % - speed is the address in response buffer where the read speed (DATA8) should be 
            %   stored
            % - tacho is the address in response buffer where the read tacho (DATA32) should be
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opOUTPUT_READ,LC0(layer),LC0(no),GV0(speed),GV0(tacho)
            %
            % Example::
            %           dataStream.opOUTPUT_READ(0,MotorPort.MotorA,0,1)
            
            dataStream.addDirectDataStream(ByteCodes.OutputRead);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(tacho);
            dataStream.GV0(speed);
        end 
        
        function opOUTPUT_TEST(dataStream,layer,nos,busy)
            % DataStream.opOUTPUT_TEST Add a opOUTPUT_TEST
            %
            % DataStream.opOUTPUT_READ(layer,nos) adds a opOUTPUT_TEST opcode 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - busy is the address in response buffer where the busy flag (DATA8) should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opOUTPUT_TEST,LC0(layer),LC0(nos),GV0(value)
            %
            % Example::
            %           dataStream.opOUTPUT_TEST(0,MotorBitfield.MotorA,0)
            
            dataStream.addDirectDataStream(ByteCodes.OutputTest);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.GV0(busy);
        end
        
        function opOUTPUT_READY(dataStream,layer,nos)
            % DataStream.opOUTPUT_READY Add a opOUTPUT_READY
            %
            % DataStream.opOUTPUT_READY(layer,nos) adds a opOUTPUT_READY opcode with 
            % to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - opOUTPUT_READY,LC0(layer),LC0(nos)
            %
            % Example::
            %           dataStream.opOUTPUT_READY(0,MotorBitfield.MotorA)
            
            dataStream.addDirectDataStream(ByteCodes.OutputReady);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
        end 
        
        function opOUTPUT_STEP_POWER(dataStream,layer,nos,power,step1,step2,step3,brake)
            % DataStream.opOUTPUT_STEP_POWER Add a opOUTPUT_STEP_POWER
            %
            % DataStream.opOUTPUT_STEP_POWER(layer,nos,power,step1,step2,step3,brake) 
            % adds a opOUTPUT_STEP_POWER opcode to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - power is the output power with [+-0..100%] range
            % - step1 is the steps used to ramp up
            % - step2 is the steps used for constant power
            % - step3 is the steps used for ramp down
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_STEP_POWER,LC0(layer),LC0(nos),LC1(power),LC4(step1),LC4(step2),LC4(step3),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_STEP_POWER(0,MotorBitfield.MotorA,50,50,360,50,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputStepPower);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(power);
            dataStream.LC4(step1);
            dataStream.LC4(step2);
            dataStream.LC4(step3);
            dataStream.LC0(brake);
        end
        
        % Implemented @ MMI 
        function opOUTPUT_TIME_POWER(dataStream,layer,nos,power,step1,step2,step3,brake)
            % DataStream.opOUTPUT_TIME_POWER Add a opOUTPUT_TIME_POWER
            %
            % DataStream.opOUTPUT_TIME_POWER(layer,nos,power,step1,step2,step3,brake) 
            % adds a opOUTPUT_TIME_POWER opcode to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - power is the output power with [+-0..100%] range
            % - step1 is the time used to ramp up
            % - step2 is the time used for constant power
            % - step3 is the time used for ramp down
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_TIME_POWER,LC0(layer),LC0(nos),LC1(power),LC4(step1),LC4(step2),LC4(step3),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_TIME_POWER(0,MotorBitfield.MotorA,50,50,360,50,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputTimePower);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(power);
            dataStream.LC4(step1);
            dataStream.LC4(step2);
            dataStream.LC4(step3);
            dataStream.LC0(brake);          
        end
        
        function opOUTPUT_STEP_SPEED(dataStream,layer,nos,speed,step1,step2,step3,brake)
            % DataStream.opOUTPUT_STEP_SPEED Add a opOUTPUT_STEP_SPEED
            % 
            % DataStream.opOUTPUT_STEP_SPEED(layer,nos,speed,step1,step2,step3,brake) 
            % adds a opOUTPUT_STEP_SPEED opcode to the DataStream object.
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - speed is the output speed with [+-0..100%] range
            % - step1 is the steps used to ramp up
            % - step2 is the steps used for constant speed
            % - step3 is the steps used for ramp down
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_STEP_POWER,LC0(layer),LC0(nos),LC1(speed),LC4(step1),LC4(step2),LC4(step3),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_STEP_SPEED(0,MotorBitfield.MotorA,50,50,360,50,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputStepSpeed);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(speed);
            dataStream.LC4(step1);
            dataStream.LC4(step2);
            dataStream.LC4(step3);
            dataStream.LC0(brake);
        end
        
        function opOUTPUT_TIME_SPEED(dataStream,layer,nos,speed,step1,step2,step3,brake)
            % DataStream.opOUTPUT_TIME_SPEED Add a opOUTPUT_TIME_SPEED
            %
            % DataStream.opOUTPUT_TIME_SPEED(layer,nos,speed,step1,step2,step3,brake) 
            % adds a opOUTPUT_TIME_SPEED opcode to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - speed is the output speed with [+-0..100%] range
            % - step1 is the time used for ramp up
            % - step2 is the time used for constant speed
            % - step3 is the time used for ramp down
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_TIME_SPEED,LC0(layer),LC0(nos),LC1(speed),LC4(step1),LC4(step2),LC4(step3),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_TIME_SPEED(0,MotorBitfield.MotorA,50,100,1000,100,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputTimeSpeed);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(speed);
            dataStream.LC4(step1);
            dataStream.LC4(step2);
            dataStream.LC4(step3);
            dataStream.LC0(brake);
        end
        
        function opOUTPUT_STEP_SYNC(dataStream,layer,nos,power,turn,step,brake)
            % DataStream.opOUTPUT_STEP_SYNC Add a opOUTPUT_STEP_SYNC
            %
            % DataStream.opOUTPUT_STEP_SYNC(layer,nos,power,turn,step,brake) 
            % adds a opOUTPUT_STEP_SYNC opcode to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - power is the output speed with [+-0..100%] range
            % - turn is the turn ratio with [+-0..200] range
            % - step is the number of tacho pulses
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_STEP_SYNC,LC0(layer),LC0(nos),LC1(power),LC2(turn),LC4(step),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_STEP_SYNC(0,MotorBitfield.MotorA,50,30,100,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputStepSync);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(power);
            dataStream.LC2(turn);
            dataStream.LC4(step);
            dataStream.LC0(brake);
        end
        
        function opOUTPUT_TIME_SYNC(dataStream,layer,nos,power,turn,time,brake)
            % DataStream.opOUTPUT_TIME_SYNC Add a opOUTPUT_TIME_SYNC
            %
            % DataStream.opOUTPUT_TIME_SYNC(layer,nos,power,turn,time,brake) 
            % adds a opOUTPUT_TIME_SYNC opcode to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - power is the output power with [+-0..100%] range
            % - turn is the turn ratio with [+-0..200] range
            % - time is the time in ms
            % - brake is [0..1] (0=Coast,  1=Brake)
            % - opOUTPUT_STEP_SYNC,LC0(layer),LC0(nos),LC1(power),LC2(turn),LC4(time),LC0(brake)
            %
            % Example::
            %           dataStream.opOUTPUT_TIME_SYNC(0,MotorBitfield.MotorA,50,30,1000,BrakeMode.Coast)
            
            dataStream.addDirectDataStream(ByteCodes.OutputTimeSync);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
            dataStream.LC1(power);
            dataStream.LC2(turn);
            dataStream.LC4(time);
            dataStream.LC0(brake);
        end
        
        function opOUTPUT_CLR_COUNT(dataStream,layer,nos)
            % DataStream.opOUTPUT_CLR_COUNT Add a opOUTPUT_CLR_COUNT
            %
            % DataStream.opOUTPUT_TIME_SYNC(layer,nos) adds a opOUTPUT_CLR_COUNT opcode 
            % to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NOS is a bit field representing output 1 to 4 (0x01, 0x02, 0x04, 0x08)
            % - opOUTPUT_CLR_COUNT,LC0(layer),LC0(nos)
            %
            % Example::
            %           dataStream.opOUTPUT_CLR_COUNT(0,MotorBitfield.MotorA)

            dataStream.addDirectDataStream(ByteCodes.OutputClrCount);
            dataStream.LC0(layer);
            dataStream.LC0(nos);
        end
        
        function opOUTPUT_GET_COUNT(dataStream,layer,no,tacho)
            % DataStream.opOUTPUT_GET_COUNT Add a opOUTPUT_GET_COUNT
            %
            % DataStream.opOUTPUT_GET_COUNT(layer,no,tacho) adds a opOUTPUT_GET_COUNT
            % to the DataStream object. 
            %
            % Notes::
            % - layer is the usb chain layer (usually 0)
            % - NO is output motor number [0...3] ( != NOS)
            % - tacho is the address in response buffer where read tacho (DATA32) should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opOUTPUT_GET_COUNT,LC0(layer),LC0(NO),GV0(tacho)
            %
            % Example::
            %           dataStream.opOUTPUT_GET_COUNT(0,MotorPort.MotorA,4)
 
            dataStream.addDirectDataStream(ByteCodes.OutputGetCount);
            dataStream.LC0(layer);
            dataStream.LC0(no);
            dataStream.GV0(tacho);
        end
        
        % Implemented @ MMI
        function opCOM_TEST(dataStream,hardware,name,busy)
            % DataStream.opCOM_TEST Add a opCOM_TEST
            %
            % DataStream.opCOM_TEST(hardware,name,value) adds a opCOM_TEST to
            % the DataStream object.
            %
            % Notes::
            % - hardware is the communication adapter to be tested. 
            %   -> 1: USB, 2: BT, 3: Wifi
            % - name is the name of the device ('0': own adapter) (?)
            % - busy is the address in response buffer where the busy flag (DATA8) should be 
            %   stored
            %   -> more info about response buffer: refer to GV0(dataStream,i) 
            % - opCOM_TEST,LC0(hardware),LCS,'0',GV0(value)
            %
            % Example::
            %           dataStream.opCOM_TEST(1,'0',0)
            
            dataStream.addDirectDataStream(ByteCodes.COMTest);
            dataStream.LC0(hardware);
            dataStream.addLCSString(name); 
            dataStream.GV0(value);
        end
        
        % Implemented @ MMI
        function opCOM_READY(dataStream,hardware,name)
            % DataStream.opCOM_READY Add a opCOM_READY
            %
            % DataStream.opCOM_READY(hardware,name,value) adds a opCOM_READY to
            % the DataStream object.
            %
            % Notes::
            % - hardware is the communication adapter to be tested. 
            %   -> 1: USB, 2: BT, 3: Wifi
            % - name is the name of the device ('0': own adapter) (?)
            % - state is 0 when ready and 1 when busy. (DATA8)
            % - opCOM_READY,LC0(hardware),LCS,'0'
            %
            % Example::
            %           dataStream.opCOM_TEST(1,'0')
            
            dataStream.addDirectDataStream(Bytecodes.COMReady);
            dataStream.LC0(hardware);
            dataStream.addLCSString(name);
        end
        
        function opCOMGET_GET_BRICKNAME(dataStream,length,name)
            % DataStream.opCOMGET_GET_BRICKNAME Add a opCOMGET_GET_BRICKNAME
            %
            % DataStream.opCOMGET_GET_BRICKNAME(length,name) adds a opCOMGET
            % with a GET_BRICKNAME subcode to the DataStream object. 
            %
            % Notes::
            % - length is the maximal length of string returned (-1 no check)
            % - name is the address in response buffer where string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opCOMGET_GET_BRICKNAME,LC0(13),LC0(length),GV0(name)
            %
            % Example::
            %           dataStream.opCOMGET_GET_BRICKNAME(10,0)
 
            dataStream.addDirectDataStream(ByteCodes.COMGet);
            dataStream.LC0(COMGetSubCodes.GetBrickName);
            dataStream.LC0(length);
            dataStream.GV0(name);
        end 
        
        function opCOMSET_SET_BRICKNAME(dataStream,name)
            % DataStream.opCOMSET_SET_BRICKNAME Add a opCOMSET_SET_BRICKNAME
            %
            % DataStream.opCOMSET_SET_BRICKNAME(name) adds a opCOMSET
            % with a SET_BRICKNAME subcode to the DataStream object. 
            %
            % Notes::
            % - name is the brick name to be set
            % - opCOMSET_SET_BRICKNAME,LC0(13),LCS,'E','V','3',0
            %
            % Example::
            %           dataStream.opCOMSET_SET_BRICKNAME('EV3')
 
            dataStream.addDirectDataStream(ByteCodes.COMSet);
            dataStream.LC0(COMSetSubCodes.SetBrickName);
            dataStream.addLCSString(name);  
        end
        
        function opCOMGET_NETWORK(dataStream,hardware,length,name,mac,ip)
            % DataStream.opCOMGET_NETWORK Add a opCOMGET_NETWORK
            %
            % DataStream.opCOMGET_NETWORK(length,name) adds a opCOMGET
            % with a GET_NETWORK subcode to the DataStream object. 
            %
            % Notes::
            % - hardware is the hardware type (should always be 3 for this method)
            % - length is the maximal length of string returned (-1 no check)
            % - name is the address in response buffer where name-string should begin
            % - mac is the address in response buffer where mac-string should begin
            % - ip is the address in response buffer where ip-string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opCOMGET_GET_BRICKNAME,LC0(13),LC0(hardware),LC0(length)
            %
            % Example::
            %           dataStream.opCOMGET_NETWORK(3,18,0,4,10)
            
            dataStream.addDirectDataStream(ByteCodes.COMGet);
            dataStream.LC0(COMGetSubCodes.GetNetwork);
            dataStream.LC0(hardware);
            dataStream.LC0(length);
            dataStream.GV0(name);
            dataStream.GV0(mac);
            dataStream.GV0(ip);
        end
        
        function opCOMGET_ID(dataStream,hardware,length,id)
            % DataStream.opCOMGET_ID Add a opCOMGET_ID
            %
            % DataStream.opCOMGET_ID(hardware,length,id) adds a opCOMGET
            % with a GET_ID subcode to the DataStream object. 
            %
            % Notes::
            % - hardware is the hardware type (should always be 2 for this method)
            % - length is the maximal length of string returned (-1 no check)
            % - id is the address in response buffer where string should begin
            %   -> more info about response buffer: refer to GV0(dataStream,i)
            % - opCOMGET_ID,LC0(13),LC0(hardware),LC0(length),GV0(id)
            %
            % Example::
            %           dataStream.opCOMGET_ID(1,20,0)
 
            dataStream.addDirectDataStream(ByteCodes.COMGet);
            dataStream.LC0(COMGetSubCodes.GetID);
            dataStream.LC0(hardware);
            dataStream.LC0(length);
            dataStream.GV0(id);
        end
        
        function opMAILBOX_WRITE(dataStream,brickname,boxname,type,msg)
            % DataStream.opMAILBOX_WRITE Add a opMAILBOX_WRITE
            %
            % DataStream.opMAILBOX_WRITE(brickname,boxname,type,msg) adds a
            % opMAILBOX_WRITE to the DataStream object.
            %
            % Notes::
            % - brickname is the name of remote device
            % - boxname is the name of the receiving mailbox
            % - type is the data type of the values being sent where DATA_8
            % (0) is logic, DATA_F (3) is numeric and DATA_S (4) is text
            % - msg is the message to be sent
            % - opMAILBOX_WRITE,LCS,'T','5','0','0',0,0,LCS,'a','b','c',LC0(0),LC0(1),LC0(1)
            %
            % Example::
            %           dataStream.opMAILBOX_WRITE('T500','abc','logical',1)
            
            dataStream.addDirectDataStream(ByteCodes.MailboxWrite);
            dataStream.addLCSString(brickname);
            % hardware transportation media (not used)
            dataStream.LC0(0);
            dataStream.addLCSString(boxname);
            % type
            switch type
                case 'logic'
                    dataStream.LC0(0);
                    dataStream.LC0(1);
                    dataStream.LC0(msg);
                case 'numeric'
                    dataStream.LC0(3);
                    dataStream.LC0(1);
                    dataStream.LC4(typecast(single(msg),'int32'));
                case 'text'
                    dataStream.LC0(4);
                    dataStream.LC0(1);
                    dataStream.addLCSString(msg);
                 otherwise
                    fprintf('Error! Type must be ''text'', ''numeric'' or ''logic''.\n');
            end
        end
    
        function BEGIN_DOWNLOAD(dataStream,filelength,filename)
            % DataStream.BEGIN_DOWNLOAD Add a BEGIN_DOWNLOAD 
            %
            % DataStream.BEGIN_DOWNLOAD(filelength,filename) adds a BEGIN_DOWNLOAD
            % system DataStream to the DataStream object. Download is from PC to
            % brick.
            %
            % Notes::
            % - filelength is the length of the file in bytes
            % - filename is the PC file for download to the brick relative
            % to '/home/root/lms2012/sys' directory
            % - ss (BEGIN_DOWNLOAD) llllllll (filelength) nn.. (filename)
            %
            % Example::
            %           dataStream.BEGIN_DOWNLOAD(60,'../apps/tst/tst.rbf')
            
            dataStream.addSystemDataStream(SystemDataStreams.BeginDownload);
            dataStream.addArray(typecast(uint32(filelength),'uint8'));
            dataStream.addString(filename);            
        end
            
        function CONTINUE_DOWNLOAD(dataStream,handle,payload)
            % DataStream.CONTINUE_DOWNLOAD Add a CONTINUE_DOWNLOAD
            %
            % DataStream.CONTINUE_DOWNLOAD(handle,payload) adds a
            % CONTINUE_DOWNLOAD system DataStream to the DataStream object.
            % Download is from PC to brick.
            %
            % Notes::
            % - handle is the handle returned from BEGIN_DOWNLOAD
            % - payload is the byte data to be downloaded to the brick
            % - ss (CONTINUE_DOWNLOAD) hh (handle from BEGIN_DOWNLOAD REPLY) pppppppp ... (payload)
            %
            % Example::
            %           dataStream.CONTINUE_DOWNLOAD(0,[10 20 30 .... ])
            
            dataStream.addSystemDataStream(SystemDataStreams.ContinueDownload);
            dataStream.addValue(handle);
            dataStream.addArray(payload);
        end
        
        function BEGIN_UPLOAD(dataStream,filelength,filename)
            % DataStream.BEGIN_UPLOAD Add a BEGIN_UPLOAD
            %
            % DataStream.BEGIN_UPLOAD(filelength,filename) adds a BEGIN_UPLOAD
            % system DataStream to the DataStream object. Upload is from brick to
            % PC.
            %
            % Notes::
            % - filelength is the max buffer size used for file upload
            % - filename is the file to be uploaded from brick to PC
            % relative to '/home/root/lms2012/sys' directory
            % - ss (BEGIN_UPLOAD) llll (filelength/bytes to read) nn..(filename including path)
            %
            % Example::
            %            dataStream.BEGIN_UPLOAD(100,'../apps/tst/tst.rbf')
           
           dataStream.addSystemDataStream(SystemDataStreams.BeginUpload);
           dataStream.addArray(typecast(uint16(filelength),'uint8'));
           dataStream.addString(filename);
        end
        
        function CONTINUE_UPLOAD(dataStream,handle,maxlength)
            % DataStream.CONTINUE_UPLOAD Add a CONTINUE_UPLOAD
            %
            % DataStream.CONTINUE_UPLOAD(handle,maxlength) adds a
            % CONTINUE_UPLOAD system DataStream to the commanc object. Upload
            % is from brick to PC.
            %
            % Notes::
            % - handle is the handle returned from BEGIN_UPLOAD
            % - maxlength is the max buffer size used for file upload
            % - ss (CONTINUE_UPLOAD) hh (handle from BEGIN_UPLOAD REPLY) llll (maxlength/bytes to read)
            % Example::
            %           dataStream.CONTINUE_UPLOAD(0,100)
            
            dataStream.addSystemDataStream(SystemDataStreams.ContinueUpload);
            dataStream.addValue(handle);
            dataStream.addArray(typecast(uint16(maxlength),'uint8'));
        end
        
        function LIST_FILES(dataStream,maxlength,pathname)
            % DataStream.LIST_FILES Add a LIST_FILES
            %
            % DataStream.LIST_FILES(maxlength,pathname) adds a LIST_FILES
            % system DataStream to the DataStream object.
            %
            % Notes::
            % - maxlength is the max buffer size used for file listing
            % - pathname is the absolute pathname use for file listing
            % - ss (LIST_FILES) llll (maxlength/max bytes to read) nn..(pathname)
            %
            % Example::
            %           dataStream.LIST_FILES(100,'/home/root/lms2012/')
            
            dataStream.addSystemDataStream(SystemDataStreams.ListFiles);
            dataStream.addArray(typecast(uint16(maxlength),'uint8'));
            dataStream.addString(pathname)
        end
        
        function CONTINUE_LIST_FILES(dataStream,handle,maxlength)
            % DataStream.CONTINUE_LIST_FILES Add a CONTINUE_LIST_FILES
            %
            % DataStream.CONTINUE_LIST_FILES(handle,maxlength) adds a
            % CONTINUE_LIST_FILES system DataStream to the DataStream object.
            %
            % Notes::
            % - handle is the handle returned by LIST_FILES
            % - maxlength is the max buffer size used for file listing
            % - ss (CONTINUE_LIST_FILES) hh (handle) llll (maxlength/max bytes to read)
            
            %
            % Example::
            %           dataStream.CONTINUE_LIST_FILES(0,100)
            
            dataStream.addSystemDataStream(SystemDataStreams.ContinueListFiles);
            dataStream.addValue(handle);
            dataStream.addArray(typecast(uint16(maxlength),'uint8'));
        end
        
        function CREATE_DIR(dataStream,pathname)
            % DataStream.CREATE_DIR Add a CREATE_DIR
            %
            % DataStream.CREATE_DIR(pathname) adds a CREATE_DIR system DataStream
            % to the DataStream object.
            %
            % Notes::
            % - pathname is the absolute path for directory creation.
            % - ss (CREATE_DIR) pp (pathname)
            
            %
            % Example::
            %           dataStream.CREATE_DIR('/home/root/lms2012/newdir')
            
            dataStream.addSystemDataStream(SystemDataStreams.CreateDir);
            dataStream.addString(pathname);
        end
        
        function DELETE_FILE(dataStream,pathname)
            % DataStream.DELETE_FILE Add a DELETE_FILE
            %
            % DataStream.DELETE_FILE(pathname) adds a DELETE_FILE system
            % DataStream to the DataStream object.
            %
            % Notes::
            % - pathname is the absolute file path for deleteion
            % - ss (DELETE_FILE) pp (pathname)
            %
            % Example::
            %           dataStream.DELETE_FILE('/home/root/lms2012/newdir')
            
            dataStream.addSystemDataStream(SystemDataStreams.DeleteFile);
            dataStream.addString(pathname);
        end
        
        function WRITEMAILBOX(dataStream,title,type,msg)
            % DataStream.WRITEMAILBOX Add a WRITEMAILBOX
            %
            % DataStream.WRITEMAILBOX(title,type,msg) adds a WRITEMAILBOX
            % DataStream to the DataStream object. 
            %
            % Notes::
            % - title is the message title sent from the brick
            % - type is the sent message type being either 'text',
            % 'numeric' or 'logic'
            % - msg is the message to be sent
            %
            % Example::
            %           dataStream.WRITEMAILBOX('abc','text','hello!')
 
            dataStream.addSystemDataStream(SystemDataStreams.WriteMailBox);
            dataStream.addValue(length(title)+1);
            dataStream.addString(title);
            switch type
                case 'text'
                    dataStream.addArray(typecast(uint16(length(msg)+1),'uint8'));
                    dataStream.addString(msg);
                case 'numeric'
                    dataStream.addArray(typecast(uint16(4),'uint8'));
                    dataStream.addArray(typecast(single(msg),'uint8'));
                case 'logic'
                    dataStream.addArray(typecast(uint16(1),'uint8'));
                    if msg(1) >= 1
                        msg(1) = 1;
                    else msg(1) = 0;
                    end
                    dataStream.addArray(typecast(uint8(msg),'uint8'));
            end
        end 
        
    end
end
