classdef Program < DataStream
    properties
    end
    
    methods
        function addHeader(packet, varargin)
            
        end
        
        function PROGRAMHeader(cmd,VersionInfo,NumberOfObjects,GlobalBytes)
            % Command.PROGRAMHeader add a PROGRAMHeader
            %
            % Command.PROGRAMHeader(VersionInfo,NumberOfObjects,GlobalBytes) 
            % adds a PROGRAMHeader to the command object.
            %
            % Notes::
            % - VersionInfo is not used in the current byte implementation
            % (hardcoded to 1.04)
            % - NumberOfObjects is the number of objects used
            % - GlobalBytes is the number of global bytes
            % - PROGRAMHEADER is defined as 'L','E','G','O',LONGToBytes(0),WORDToBytes((UWORD)(BYTECODE_VERSION * 100.0)),WORDToBytes(NumberOfObjects),LONGToBytes(GlobalBytes)
            %
            % Example::
            %           cmd.PROGRAMHeader(0,1,0)
            
            BYTECODE_VERSION = 1.04;
            cmd.msg = [cmd.msg,uint8('L'),uint8('E'),uint8('G'),uint8('O')] ;
            cmd.LONGToBytes(0);
            cmd.WORDToBytes(BYTECODE_VERSION*100);
            cmd.WORDToBytes(NumberOfObjects);
            cmd.LONGToBytes(GlobalBytes);
        end
        
        function VMTHREADHeader(cmd,OffsetToInstructions,LocalBytes)
            % Command.VMTHREADHeader Add a VMTHREADHeader
            %
            % Command.VMTHREADHeader(OffsetToInstructions,LocalBytes)
            % adds a VMTHREADHeader to the command object.
            %
            % Notes::
            % - OffsetToInstructions is the offset to the instructions
            % - LocalBytes is the number of local bytes
            % - VMTHREADHeader is defined as LONGToBytes(OffsetToInstructions),0,0,0,0,LONGToBytes(LocalBytes)
            %
            % Example::
            %           cmd.VMTHREADHeader(0,1)
            
            cmd.LONGToBytes(OffsetToInstructions);
            cmd.msg = [cmd.msg uint8(0),uint8(0),uint8(0),uint8(0)];
            cmd.LONGToBytes(LocalBytes) ;
        end
        
        function SUBCALLHeader(cmd,OffsetToInstructions,LocalBytes)
            % Command.SUBCALLHeader Add a SUBCALLHeader
            %
            % Command.SUBCALLHeader(OffsetToInstructions,LocalBytes)
            % adds a SUBCALLHeader to the command object.
            %
            % Notes::
            % - OffsetToInstructions is the offset to the instructions
            % - LocalBytes is the number of local bytes
            % - SUBCALLHeader is defined as LONGToBytes(OffsetToInstructions),0,0,1,0,LONGToBytes(LocalBytes)
            %
            % Example::
            %           cmd.SUBCALLHeader(0,1)
            
            cmd.LONGToBytes(OffsetToInstructions);
            cmd.msg = [cmd.msg uint8(0),uint8(0),uint8(1),uint8(0)];
            cmd.LONGToBytes(LocalBytes) ;
        end
        
        function BLOCKHeader(cmd,OffsetToInstructions,OwnerObjectId,TriggerCount)
            % Command.BLOCKHeader
            %
            % Command.BLOCKHeader(OffsetToInstructions,OwnerObjectId,TriggerCount)
            % adds a BLOCKHeader to the command object.
            %
            % Notes::
            % - OffsetToInstructions is the offset to the instructions
            % - OwnerObjectId is the owner object id
            % - TriggerCount is the trigger count
            % - BLOCKHeader is defined as LONGToBytes(OffsetToInstructions),WORDToBytes(OwnerObjectId),WORDToBytes(TriggerCount),LONGToBytes(0)
            %
            % Example::
            %           cmd.BLOCKHeader(0,0,0)
            
            cmd.LONGToBytes(OffsetToInstructions);
            cmd.WORDToBytes(OwnerObjectId);
            cmd.WORDToBytes(TriggerCount)
            cmd.LONGToBytes(0);
        end  
        
        function addLength(cmd)
            % Command.addFileSize Add file size
            % 
            % Command.addFileSize() adds the file size to the command
            % object.
            %
            % Notes::
            % - With bytecode compiling using "Old header", the file size is 
            % inserted at byte number 5 which corresponds to the two bytes 
            % after 'L','E','G','O'
            %
            % Example::
            %            cmd.addFileSize
            
            cmd.msg(5:6) = typecast(uint16(length(cmd.msg)),'uint8');            
        end
        
        function opOBJECT_END(cmd)
            % Command.opOBJECT_END Add a opOBJECT_END
            % 
            % Command.opOBJECT_END() adds a opOBJECT_END opcode to the
            % command object. 
            %
            % Notes::
            % - opOBJECT_END
            %
            % Example::            
            %           cmd.opOBJECT_END()
           
            cmd.addDirectCommand(ByteCodes.ObjectEnd);
        end
        
        function opNOP(cmd)
            % Command.opNOP Add a opNOP 
            %
            % Command.opNOP() adds a opNOP opcode to the command object.
            %
            % Example::
            %           cmd.opNOP()
           
            cmd.addDirectCommand(ByteCodes.Nop);            
        end
        
        function opJR(cmd,offset)
            % Command.opJR Add a opJR
            %
            % Command.opJR(offset) adds a opJR opcode to the command
            % object.
            %
            % Notes::
            % - offset is the number of command to jump where the sign
            % indicates the jump direction
            % - opJR,LC0(offset)
            %
            % Example::
            %           cmd.opJR(-10)

            cmd.addDirectCommand(ByteCodes.Jr);
            cmd.LC0(offset);
        end
    end
end

